"""Utilities to extract information from langchain tools for use in prompts."""
import inspect
from textwrap import dedent
from typing import List

from langchain.tools.base import StructuredTool

from langchain_benchmarks.tool_usage.agents.experimental.encoder import (
    FunctionDefinition,
    Parameter,
)

# PUBLIC API


def get_parameters_from_tool(tool: StructuredTool) -> List[Parameter]:
    """Convert a langchain tool to a tool user tool."""
    schema = tool.args_schema.schema()

    properties = schema["properties"]
    parameters = []
    # Is this needed or is string OK?
    type_adapter = {
        "string": "str",  # str or string?
        "integer": "int",
        "number": "float",
        "boolean": "bool",
    }
    for key, value in properties.items():
        parameters.append(
            {
                "name": key,
                "type": type_adapter.get(value["type"], value["type"]),
                "description": value.get("description", ""),
            }
        )

    return parameters


#
def convert_tool_to_function_definition(tool: StructuredTool) -> FunctionDefinition:
    """Convert a langchain tool to a tool user tool."""
    # Here we re-inspect the underlying function to get the doc-string
    # since StructuredTool modifies it, but we want the raw one for maximum
    # flexibility.
    description = inspect.getdoc(tool.func)

    parameters = get_parameters_from_tool(tool)
    return {
        "name": tool.name,
        "description": dedent(description),
        "parameters": parameters,
        "return_value": {
            "type": "Any",
        },
    }



"""Factory for creating agents for the tool usage task."""
from typing import Optional

from langchain.agents import AgentExecutor
from langchain_core.runnables import Runnable, RunnableConfig

from langchain_benchmarks import RateLimiter, model_registry
from langchain_benchmarks.schema import ToolUsageTask
from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter
from langchain_benchmarks.tool_usage.agents.experimental.agent import create_agent
from langchain_benchmarks.tool_usage.agents.experimental.parser import (
    GenericAgentParser,
)


class CustomAgentFactory:
    """A factory for creating tool using agents.

    A factory for agents that do not leverage any special JSON mode for
    function usage; instead all function invocation behavior is implemented solely
    through prompt engineering and parsing.
    """

    def __init__(
        self,
        task: ToolUsageTask,
        model: str,
        *,
        rate_limiter: Optional[RateLimiter] = None,
        num_retries: int = 0,
    ) -> None:
        """Create an agent factory for the given tool usage task.

        Args:
            task: The task to create an agent factory for
            model: model name (check model_registry)
            rate_limiter: The rate limiter to use if provided
            num_retries: The number of times to retry the agent if it fails
        """
        if model not in model_registry:
            raise ValueError(f"Unknown model: {model}")
        self.task = task
        self.model = model
        self.rate_limiter = rate_limiter
        self.num_retries = num_retries

    def __call__(self) -> Runnable:
        if isinstance(self.model, str):
            registered_model = model_registry.get_model(self.model)
            if registered_model is None:
                raise ValueError(f"Unknown model: {self.model}")
            model = registered_model.get_model(model_params={"temperature": 0})
        else:
            model = self.model

        def _add_task_instructions(
            input: dict, config: Optional[RunnableConfig] = None, **kwargs
        ) -> dict:
            """Add task instructions to the question."""
            if not isinstance(input, dict):
                raise ValueError(
                    f"Expected input to be a dict with key `question`. "
                    f"Found {type(input)}."
                )
            input = input.copy()
            input["question"] = (
                f"{self.task.instructions}\nWrite down your answer, "
                f"but do not explain it. Input: `{input['question']}`"
            )
            return input

        env = self.task.create_environment()

        agent = create_agent(
            model,
            env.tools,
            GenericAgentParser(wrapping_xml_tag="tool", require_closing_xml_tag=False),
            rate_limiter=self.rate_limiter,
        )
        if self.num_retries > 0:
            agent = agent.with_retry(
                stop_after_attempt=self.num_retries + 1,
            )
        executor = AgentExecutor(
            agent=agent,
            tools=env.tools,
            handle_parsing_errors=True,
            return_intermediate_steps=True,
        )

        return _add_task_instructions | apply_agent_executor_adapter(
            executor, state_reader=env.read_state
        )



from __future__ import annotations

from langchain_benchmarks.schema import ModelRegistry, RegisteredModel

_OPEN_AI_MODELS = [
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-1106",
        type="chat",
        description=(
            "The latest GPT-3.5 Turbo model with improved instruction following, "
            "JSON mode, reproducible outputs, parallel function calling, and more. "
            "Returns a maximum of 4,096 output tokens."
        ),
        params={
            "model": "gpt-3.5-turbo-1106",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo",
        type="chat",
        description="Currently points to gpt-3.5-turbo-0613.",
        params={
            "model": "gpt-3.5-turbo",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-16k",
        type="chat",
        description="Currently points to gpt-3.5-turbo-0613.",
        params={
            "model": "gpt-3.5-turbo-16k",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-instruct",
        type="llm",
        description=(
            "Similar capabilities as text-davinci-003 but compatible with legacy "
            "Completions endpoint and not Chat Completions."
        ),
        params={
            "model": "gpt-3.5-turbo-instruct",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-0613",
        type="chat",
        description=(
            "Legacy Snapshot of gpt-3.5-turbo from June 13th 2023. "
            "Will be deprecated on June 13, 2024."
        ),
        params={
            "model": "gpt-3.5-turbo-0613",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-16k-0613",
        type="chat",
        description=(
            "Legacy Snapshot of gpt-3.5-16k-turbo from June 13th 2023. "
            "Will be deprecated on June 13, 2024."
        ),
        params={
            "model": "gpt-3.5-turbo-16k-0613",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-3.5-turbo-0301",
        type="chat",
        description=(
            "Legacy Snapshot of gpt-3.5-turbo from March 1st 2023. "
            "Will be deprecated on June 13th 2024."
        ),
        params={
            "model": "gpt-3.5-turbo-0301",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="text-davinci-003",
        type="llm",
        description=(
            "Legacy Can do language tasks with better quality and consistency than "
            "the curie, babbage, or ada models. Will be deprecated on Jan 4th 2024."
        ),
        params={
            "model": "text-davinci-003",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="text-davinci-002",
        type="llm",
        description=(
            "Legacy Similar capabilities to text-davinci-003 but trained with "
            "supervised fine-tuning instead of reinforcement learning. "
            "Will be deprecated on Jan 4th 2024."
        ),
        params={
            "model": "text-davinci-002",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="code-davinci-002",
        type="llm",
        description="Legacy Optimized for code-completion tasks. Will be deprecated "
        "on Jan 4th 2024.",
        params={
            "model": "code-davinci-002",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-4-1106-preview",
        type="chat",
        description="GPT-4 TurboNew - The latest GPT-4 model with improved instruction following, JSON mode, reproducible outputs, parallel function calling, and more. Returns a maximum of 4,096 output tokens. This preview model is not yet suited for production traffic.",
        params={
            "model": "gpt-4-1106-preview",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-4-0613",
        type="chat",
        description="Snapshot of gpt-4 from June 13th 2023 with improved function calling support.",
        params={
            "model": "gpt-4-0613",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-4-32k-0613",
        type="chat",
        description="Snapshot of gpt-4-32k from June 13th 2023 with improved function calling support.",
        params={
            "model": "gpt-4-32k-0613",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-4-0314",
        description="Snapshot of gpt-4 from March 14th 2023 with function calling support. This model version will be deprecated on June 13th 2024.",
        type="chat",
        params={
            "model": "gpt-4-0314",
        },
    ),
    RegisteredModel(
        provider="openai",
        name="gpt-4-32k-0314",
        description="Snapshot of gpt-4-32k from March 14th 2023 with function calling support. This model version will be deprecated on June 13th 2024.",
        type="chat",
        params={
            "model": "gpt-4-32k-0314",
        },
    ),
]

_FIREWORKS_MODELS = [
    RegisteredModel(
        provider="fireworks",
        name="llama-v2-7b-chat-fw",
        type="chat",
        description="7b parameter LlamaChat model",
        params={
            "model": "accounts/fireworks/models/llama-v2-7b-chat",
        },
    ),
    RegisteredModel(
        provider="fireworks",
        name="llama-v2-13b-chat-fw",
        type="chat",
        description="13b parameter LlamaChat model",
        params={
            "model": "accounts/fireworks/models/llama-v2-13b-chat",
        },
    ),
    RegisteredModel(
        provider="fireworks",
        name="llama-v2-70b-chat-fw",
        type="chat",
        description="70b parameter LlamaChat model",
        params={
            "model": "accounts/fireworks/models/llama-v2-70b-chat",
        },
    ),
    RegisteredModel(
        provider="fireworks",
        name="yi-34b-200k-fw",
        type="llm",
        description=" 4B LLM model from 01.ai, with context window 200k.",
        params={
            "model": "accounts/fireworks/models/yi-34b-200k",
        },
    ),
    RegisteredModel(
        provider="fireworks",
        name="mixtral-8x7b-instruct-fw",
        description="Mistral MoE 8x7B Instruct v0.1 model with Sparse "
        "Mixture of Experts. Fine tuned for instruction following",
        type="llm",
        params={"model": "accounts/fireworks/models/mixtral-8x7b-instruct"},
    ),
]

_ANTHROPIC_MODELS = [
    RegisteredModel(
        provider="anthropic",
        name="claude-2",
        description=("Superior performance on tasks that require complex reasoning"),
        type="chat",
        params={
            "model": "claude-2",
        },
    ),
    RegisteredModel(
        provider="anthropic",
        name="claude-2.1",
        description=(
            "Same performance as Claude 2, plus significant reduction in model "
            "hallucination rates"
        ),
        type="chat",
        params={
            "model": "claude-2.1",
        },
    ),
    RegisteredModel(
        provider="anthropic",
        name="claude-instant-1.2",
        description="low-latency, high throughput.",
        type="chat",
        params={
            "model": "claude-instant-1.2",
        },
    ),
    RegisteredModel(
        provider="anthropic",
        name="claude-instant-1",
        description="low-latency, high throughput.",
        type="chat",
        params={
            "model": "claude-instant-1",
        },
    ),
]
_GOOGLE_GENAI_MODELS = [
    RegisteredModel(
        provider="google-genai",
        name="gemini-pro",
        description="Gemini Pro is a large model from Google trained on a diverse set of tasks.",
        type="chat",
        params={
            "model": "gemini-pro",
            "convert_system_message_to_human": True,
        },
    )
]

_ANYSCALE_MODELS = [
    RegisteredModel(
        provider="anyscale",
        name="mistral-7b-instruct-v0.1",
        description="Mistral 7B model fine-tuned for function-calling.",
        type="chat",
        params={
            "model": "mistralai/Mistral-7B-Instruct-v0.1",
        },
    ),
]

model_registry = ModelRegistry(
    registered_models=_OPEN_AI_MODELS
    + _FIREWORKS_MODELS
    + _ANYSCALE_MODELS
    + _ANTHROPIC_MODELS
    + _GOOGLE_GENAI_MODELS
)



import ast
import re
from typing import Dict, Optional, Union

from langchain.agents import AgentOutputParser
from langchain.pydantic_v1 import BaseModel, Field
from langchain_core.agents import AgentAction, AgentActionMessageLog, AgentFinish
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import AIMessage


class _ToolInvocationRequest(BaseModel):
    """Light-weight pydantic model for validating the raw tool invocation request.

    The purpose of this model, is to make sure that whatever as parsed from
    the raw llm output has `tool_name` and potential `arguments` fields, and
    nothing else.
    """

    tool_name: str
    # OK parameterless tools which do not take arguments
    arguments: Optional[Dict] = Field(default_factory=dict)


class GenericAgentParser(AgentOutputParser):
    """A generalized parser that makes it easier to parameterize different parsing."""

    wrapping_xml_tag: str
    """The tag that wraps the function invocation request.
    
    For example, if "tool", then the function invocation request should be wrapped
    in <tool>...</tool>.
    """
    require_closing_xml_tag: bool = False
    """Whether we should require a closing tag for the wrapping_xml_tag.
    
    For example, if True, then the function invocation request should be wrapped
    """

    def parse(self, text: str) -> Union[AgentFinish, AgentAction]:
        """Parse the output of the agent."""
        open_tag = f"<{self.wrapping_xml_tag}>"
        close_tag = f"</{self.wrapping_xml_tag}>"
        if open_tag in text:
            # This is a hack to make sure that </tool> is always present
            # in the output if <tool>. </tool> may be a stop sequence for the
            # language model, so depending on implementation
            # the stop sequence may be cut off.
            # There might be a better way to do this, but this works and
            # is simple.
            if not self.require_closing_xml_tag:
                text += close_tag

        pattern = rf"{open_tag}(?P<invocation>.*?){close_tag}"
        match = re.search(pattern, text, re.DOTALL)
        if match:
            content = match.group("invocation").strip()
            return parse_invocation(content, self.wrapping_xml_tag)

        return AgentFinish(
            log=text,
            return_values={
                "output": text,
            },
        )


def parse_invocation(text: str, tag: str) -> AgentAction:
    """Parse the content of the function invocation.

    Args:
        text: The text to parse.
        tag: The tag that wraps the function invocation request.

    Returns:
        An AgentAction that corresponds to the function invocation.

    Raises:
        OutputParserException: If the parsing fails.

        This exception is meant to be caught by the agent executor and
        handled appropriately to provide feedback to the LLM.
    """
    ai_content = f"<{tag}>{text}</{tag}>\n"

    try:
        result = ast.literal_eval(text)
    except BaseException as e:
        # Convert this to something controllable by the user.
        err_msg = (
            f"ERROR: Please use the format "
            f'<{tag}>{{"tool_name": $TOOL_NAME, "arguments": $ARGUMENTS}}</{tag}>\n'
        )

        raise OutputParserException(
            error=e,
            llm_output=ai_content,
            observation=err_msg,
            send_to_llm=True,
        )

    try:
        request = _ToolInvocationRequest.validate(result)
    except Exception as e:  # Using broad exception since it's not just ValidationError
        # Can also raise DictError if result is not a dict.
        err_msg = (
            f"ERROR: Please use the format "
            f'<{tag}>{{"tool_name": $TOOL_NAME, "arguments": $ARGUMENTS}}</{tag}>\n'
        )
        raise OutputParserException(
            error=e,
            llm_output=ai_content,
            send_to_llm=True,
            observation=err_msg,
        )

    return AgentActionMessageLog(
        message_log=[AIMessage(content=ai_content)],
        tool=request.tool_name,
        tool_input=request.arguments,
        log=f"\nInvoking {request.tool_name}: {request.arguments}\n\t",
    )



from langchain_benchmarks.schema import RetrievalTask

# ID of public Multi Modal Slide Decks dataset
DATASET_ID = "https://smith.langchain.com/public/40afc8e7-9d7e-44ed-8971-2cae1eb59731/d"

MULTI_MODAL_SLIDE_DECKS_TASK = RetrievalTask(
    name="Multi-modal slide decks",
    dataset_id=DATASET_ID,
    retriever_factories={},
    architecture_factories={},
    get_docs={},
    description=(
        """\
This public dataset is a work-in-progress and will be extended over time.
        
Questions and answers based on slide decks containing visual tables and charts.

Each example is composed of a question and reference answer.

Success is measured based on the accuracy of the answer relative to the reference answer.
"""  # noqa: E501
    ),
)



from typing import Any, Callable, Optional

from langchain.agents import AgentExecutor
from langchain_core.runnables import Runnable, RunnableLambda, RunnablePassthrough


def _ensure_output_exists(inputs: dict) -> dict:
    """Make sure that the output key is always present."""
    if "output" not in inputs:
        return {"output": "", **inputs}
    return inputs


def apply_agent_executor_adapter(
    agent_executor: AgentExecutor,
    *,
    state_reader: Optional[Callable[[], Any]] = None,
) -> Runnable:
    """An adapter for the agent executor to standardize its input and output.

    1) Map `question` to `input` (`question` is used in the datasets,
       but `input` is used in the agent executor)
    2) Ensure that `output` is always returned (will be set to "" if missing) --
       note that this may be relaxed after more updates in the eval config.
    3) Populate `state` key in the response of the agent with the system state
       if a state reader is provided.

    Args:
        agent_executor: the agent executor
        state_reader: A callable without parameters that if invoked will return
                      the state of the environment. Used to populate the 'state' key.

    Returns:
        a new runnable with a standardized output.
    """

    def _read_state(*args: Any, **kwargs: Any) -> Any:
        """Read the state of the environment."""
        if state_reader is not None:
            return state_reader()
        else:
            return None

    def _format_input(inputs: dict) -> dict:
        """Make sure that the input is always called `input`."""

        if "question" not in inputs:
            raise ValueError(
                "Expected 'question' to be in the inputs. Found only the following "
                f"keys {sorted(inputs.keys())}."
            )

        inputs = inputs.copy()  # Because 'question' is popped below

        if "input" not in inputs:
            return {"input": inputs.pop("question"), **inputs}
        return inputs

    runnable = (
        RunnableLambda(_format_input).with_config({"run_name": "Format Input"})
        | agent_executor
        | RunnableLambda(_ensure_output_exists).with_config(
            {"run_name": "Ensure Output"}
        )
    )

    if state_reader is not None:
        runnable = runnable | RunnablePassthrough.assign(state=_read_state).with_config(
            {"run_name": "Read Env State"}
        )
    return runnable



from typing import List, Literal, Optional, Sequence, Tuple, Union

from langchain.agents import AgentOutputParser
from langchain.prompts.chat import ChatPromptTemplate
from langchain.schema.runnable import Runnable
from langchain.tools import StructuredTool
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.language_models import BaseChatModel, BaseLanguageModel
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage
from langchain_core.prompts import MessagesPlaceholder
from typing_extensions import NotRequired, TypedDict

from langchain_benchmarks import RateLimiter
from langchain_benchmarks.rate_limiting import with_rate_limit
from langchain_benchmarks.tool_usage.agents.experimental.encoder import (
    AstPrinter,
    FunctionResult,
    TypeScriptEncoder,
    XMLEncoder,
)
from langchain_benchmarks.tool_usage.agents.experimental.prompts import (
    _AGENT_INSTRUCTIONS_BLOB_STYLE,
)
from langchain_benchmarks.tool_usage.agents.experimental.tool_utils import (
    convert_tool_to_function_definition,
)


def format_steps_for_chat(
    intermediate_steps: List[Tuple[AgentAction, str]],
    ast_printer: AstPrinter,
) -> List[BaseMessage]:
    """Format the steps."""
    messages = []
    for action, observation in intermediate_steps:
        # Action messages contains the tool invocation request from the LLM
        # Now add the result of the tool invocation.

        if action.tool == "_Exception":
            messages.append(
                AIMessage(
                    content=action.log,
                )
            )
            messages.append(
                # Tool input is the error message for the exception
                HumanMessage(content=action.tool_input)
            )
        else:
            messages.extend(action.messages)
            function_result: FunctionResult = {
                "name": action.tool,
                "error": None,
                "result": observation,
            }
            messages.append(
                HumanMessage(
                    content=ast_printer.visit_function_result(function_result),
                )
            )

    return messages


# PUBLIC API


class AgentInput(TypedDict):
    """The input to the agent."""

    input: str
    """The input to the agent."""
    intermediate_steps: List[Tuple[AgentAction, str]]
    """The intermediate steps taken by the agent."""
    examples: NotRequired[List[BaseMessage]]
    """A list of messages that can be used to form example traces."""


def create_agent(
    model: Union[BaseChatModel, BaseLanguageModel],
    tools: Sequence[StructuredTool],
    parser: AgentOutputParser,
    *,
    ast_printer: Union[AstPrinter, Literal["xml"]] = "xml",
    rate_limiter: Optional[RateLimiter] = None,
) -> Runnable[AgentInput, Union[AgentAction, AgentFinish]]:
    """Create an agent for a chat model."""
    if isinstance(ast_printer, str):
        if ast_printer == "xml":
            ast_printer_ = XMLEncoder()
        elif ast_printer == "typescript":
            ast_printer_ = TypeScriptEncoder()
        else:
            raise ValueError(f"Unknown ast printer: {ast_printer}")
    elif isinstance(ast_printer, AstPrinter):
        ast_printer_ = ast_printer
    else:
        raise TypeError(
            f"Expected AstPrinter or str, got {type(ast_printer)} for `ast_printer`"
        )

    function_definitions = [convert_tool_to_function_definition(tool) for tool in tools]
    tool_description = ast_printer_.visit_function_definitions(function_definitions)

    template = ChatPromptTemplate.from_messages(
        [
            ("system", _AGENT_INSTRUCTIONS_BLOB_STYLE),
            MessagesPlaceholder("examples"),  # Can use to add example traces
            ("human", "{input}"),
            MessagesPlaceholder("history"),
        ]
    ).partial(tool_description=tool_description)

    # For the time being, hard-coding the fact that we're using a <tool> tag.
    model = model.bind(stop=["</tool>"])

    if rate_limiter:
        # Apply a rate limiter if it was provided
        model = with_rate_limit(model, rate_limiter)

    agent = (
        {
            "input": lambda x: x["input"],
            "history": lambda x: format_steps_for_chat(
                x["intermediate_steps"], ast_printer_
            ),
            "examples": lambda x: x.get("examples", []),
        }
        | template
        | model
        | parser
    )
    return agent



from langchain_core.prompts import PromptTemplate

MATH_TEMPLATE = """\
You live in an alternate universe. Do not assume that you know anything.
You are a teacher grading a quiz.
You are given a question, the student's answer, and the true answer, and are asked to score the student answer as either CORRECT or INCORRECT.

Example Format:
QUESTION: question here
STUDENT ANSWER: student's answer here
TRUE ANSWER: true answer here
GRADE: CORRECT or INCORRECT here

Given that you live in an alternate universe the TRUE answer may be different from what you expect. That's OK!

Grade the student answers based ONLY on whether it matches the TRUE answer. Ignore differences in punctuation and phrasing between the student answer and true answer. It is OK if the student answer contains more information than the true answer, as long as it does not contain any conflicting statements. Begin! 

QUESTION: {query}
STUDENT ANSWER: {result}
TRUE ANSWER: {answer}
GRADE:"""
QA_TEMPLATE_FOR_MULTIVERSE_MATH = PromptTemplate(
    input_variables=["result", "answer"], template=MATH_TEMPLATE
)

MATH_TEMPLATE_NO_QUESTION = """\
Compare the INPUT_A and INPUT_B and determine whether the numeric result in them is the same.

If the result is the same, reply with CORRECT. If the result is different, reply with INCORRECT.

Example Format:
INPUT_A: input_a here
INPUT_B: input_b here
COMPARISON: CORRECT or INCORRECT here

Ignore differences in punctuation and phrasing between the student answer and true answer. 

Begin!

INPUT_A: {answer}
INPUT_B: {result}
COMPARISON:"""

# Version without the query
QA_TEMPLATE_FOR_MULTIVERSE_MATH_WITHOUT_QUESTION = PromptTemplate(
    input_variables=["result", "answer"], template=MATH_TEMPLATE_NO_QUESTION
)



"""Code for creating an assistant factory for evaluating tool usage tasks.

See: https://platform.openai.com/docs/assistants/how-it-works/creating-assistants
"""
from typing import Optional

from langchain.agents import AgentExecutor
from langchain.agents.openai_assistant.base import OpenAIAssistantRunnable
from langchain.schema.runnable import Runnable

from langchain_benchmarks import rate_limiting
from langchain_benchmarks.schema import ToolUsageTask
from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter


class OpenAIAssistantFactory:
    def __init__(
        self,
        task: ToolUsageTask,
        *,
        model: str,
        rate_limiter: Optional[rate_limiting.RateLimiter] = None,
        num_retries: int = 0,
    ) -> None:
        """Create an OpenAI agent factory for the given task.

        Args:
            task: The task to create an agent factory for.
            model: The model to use -- this must be an open AI model.
            rate_limiter: The rate limiter to use
            num_retries: The number of times to retry the assistant if it fails
        """
        if not isinstance(model, str):
            raise ValueError(f"Expected str for model, got {type(model)}")
        self.task = task
        tools = task.create_environment().tools
        # Stateless, so we only need to create it once
        self.agent = OpenAIAssistantRunnable.create_assistant(
            name=f"{task.name} assistant",
            instructions=self.task.instructions,
            tools=tools,
            model=model,
            as_agent=True,
        )
        self.rate_limiter = rate_limiter
        self.num_retries = num_retries

    def __call__(self) -> Runnable:
        env = self.task.create_environment()

        agent = self.agent
        if self.rate_limiter is not None:
            # Rate limited model
            agent = rate_limiting.with_rate_limit(agent, self.rate_limiter)

        def _map_key(x: dict):
            # Assistant expects the 'content' key explicitly
            return {
                "content": x["input"],
                **{k: v for k, v in x.items() if k != "input"},
            }

        agent = _map_key | self.agent
        if self.num_retries > 0:
            agent = agent.with_retry(
                stop_after_attempt=self.num_retries + 1,
            )
        runnable = AgentExecutor(
            agent=agent,
            tools=env.tools,
            handle_parsing_errors=True,
            return_intermediate_steps=True,
        )

        # Returns `state` in the output if the environment has a state reader
        # makes sure that `output` is always in the output
        return apply_agent_executor_adapter(runnable, state_reader=env.read_state)



"""Wrapper around the anthropic tool user SDK.

The anthropic tool user SDK is an alpha release so this code will likely be
changed or deleted in the future. It's here simply to make it easier to benchmark
the performance of the existing tool user SDK, to compare it with the performance
of other implementations.
"""

from importlib.util import find_spec
from typing import Any, Dict, List, Optional, Sequence

from langchain.tools import StructuredTool
from langchain_core.callbacks.manager import trace_as_chain_group
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda

from langchain_benchmarks import rate_limiting
from langchain_benchmarks.schema import ToolUsageTask
from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter


def convert_langchain_tool_to_tool_user_tool(lc_tool: StructuredTool) -> Any:
    """Convert a langchain tool to a tool user tool."""
    from tool_use_package.tools.base_tool import BaseTool

    class DynamicTool(BaseTool):
        def use_tool(self, **kwargs):
            return lc_tool(kwargs)

    schema = lc_tool.args_schema.schema()

    properties = schema["properties"]
    parameters = []
    # Is this needed or is string OK?
    type_adapter = {
        "string": "str",  # str or string?
        "integer": "int",
        "number": "float",
        "boolean": "bool",
    }
    for key, value in properties.items():
        parameters.append(
            {
                "name": key,
                "type": type_adapter.get(value["type"], value["type"]),
                "description": value.get("description", ""),
            }
        )

    return DynamicTool(lc_tool.name, lc_tool.description, parameters)


def _handle_tool_inputs(
    tool_inputs: List[Dict[str, Any]],
    tools: Sequence[StructuredTool],
    config: Optional[RunnableConfig] = None,
) -> Dict[str, Any]:
    """Handle tool inputs."""
    tool_by_name = {tool.name: tool for tool in tools}
    tool_error: Optional[str] = None
    tool_outputs = []
    for tool_input in tool_inputs:
        tool_name = tool_input["tool_name"]
        tool_arguments = tool_input["tool_arguments"]
        tool = tool_by_name[tool_name]
        try:
            tool_result = tool.invoke(tool_arguments, config=config)
        except Exception as e:  # Break on first error
            tool_error = str(e)
            tool_outputs = None
            break
        tool_outputs.append(
            {
                "tool_name": tool_name,
                "tool_result": tool_result,
            }
        )
    return {
        "role": "tool_outputs",
        "tool_outputs": tool_outputs,
        "tool_error": tool_error,
    }


def run_anthropic_agent_simple(
    tools: Sequence[StructuredTool],
    user_message: str,
    *,
    max_iterations: int = 30,
    config: Optional[RunnableConfig] = None,
    **kwargs,
) -> List[dict]:
    """Make an anthropic agent."""
    from tool_use_package.tool_user import ToolUser

    verbose = kwargs.pop("verbose", False)

    tool_user = ToolUser(
        [convert_langchain_tool_to_tool_user_tool(tool) for tool in tools], **kwargs
    )
    messages = [
        {
            "role": "human",
            "content": user_message,
            "tool_error": None,
            "tool_outputs": [],
            "tool_inputs": [],
        }
    ]
    with trace_as_chain_group(
        "Anthropic Agent Run",
        inputs={"user_message": user_message},
        callback_manager=config.get("callbacks", None) if config else None,
    ) as group_manager:
        for num_iteration in range(max_iterations):
            with trace_as_chain_group(
                f"Anthropic Agent Iteration {num_iteration}",
                inputs={"messages": messages},
                callback_manager=group_manager.parent_run_manager.get_child(),
            ) as iteration_manager:
                last_message = tool_user.use_tools(
                    messages, execution_mode="manual", verbose=verbose
                )
                new_messages = [last_message]

                if last_message["role"] == "tool_inputs":
                    tool_inputs = last_message["tool_inputs"]
                    new_message = _handle_tool_inputs(
                        tool_inputs,
                        tools,
                        config={
                            "callbacks": iteration_manager.parent_run_manager.get_child(),
                        },
                    )
                    new_messages.append(new_message)

                iteration_manager.on_chain_end(outputs=new_messages)
                messages.extend(new_messages)

                # Finally break if the last message is from the assistant
                if last_message["role"] == "assistant":
                    break
        else:
            raise ValueError("Max iterations reached")
        group_manager.on_chain_end(outputs=messages)
    return messages


def convert_messages_to_finalized_output(
    messages: List[Dict[str, Any]],
) -> Dict[str, Any]:
    """Convert the history of messages into the expected output for eval.

    This matches the agent executor output which has the following structure:

    {
        "output": "The output of the agent",
        "intermediate_steps": [
            (
                AgentAction(
                    tool="add_x_y",
                    tool_input={"x": 2.0, "y": 5.0},
                    log="Invoking tool `add_x_y` with `{'x': 2.0, 'y': 5.0}`",
                ),
                9.0,
            )
        ],
        "state": Any, # Optional key for tasks that involve manipulation of an env.
    }
    """
    if not messages:
        raise ValueError("Expected at least one message")

    last_message = messages[-1]

    if last_message["role"] != "assistant":
        raise ValueError(
            f"Expected the last message to be from the assistant. "
            f"Instead got {last_message}."
        )

    actual_steps = []

    for message in messages:
        if "role" not in message:
            raise ValueError(f"Expected role in message {message}")
        role = message["role"]

        if role == "tool_inputs":
            # Get the name of the tool used
            for tool_input in message["tool_inputs"]:
                actual_steps.append(tool_input["tool_name"])

    return {
        "output": last_message["content"],
        "actual_steps": actual_steps,
    }


def create_agent(tools: Sequence[StructuredTool]) -> RunnableLambda:
    """Create an agent."""

    def run_agent(
        input: dict, config: Optional[RunnableConfig] = None, **kwargs
    ) -> dict:
        """Run the agent."""
        messages = run_anthropic_agent_simple(
            tools, input["input"], config=config, **kwargs
        )
        return convert_messages_to_finalized_output(messages)

    return RunnableLambda(run_agent)


class AnthropicToolUserFactory:
    def __init__(
        self,
        task: ToolUsageTask,
        *,
        rate_limiter: Optional[rate_limiting.RateLimiter] = None,
    ) -> None:
        """Create an OpenAI agent factory for the given task.


        Args:
            task: The task to create an agent factory for.
            rate_limiter: The rate limiter to use
        """
        self.task = task
        self.rate_limiter = rate_limiter
        if not find_spec("tool_use_package"):
            raise ImportError(
                'Could not import "tool_use_package". Please '
                "follow instructions here to install "
                "https://github.com/anthropics/anthropic-tools/tree/main"
            )

    def __call__(self, **kwargs: Any) -> Runnable:
        env = self.task.create_environment()

        def _add_task_instructions(
            input: dict, config: Optional[RunnableConfig] = None, **kwargs
        ) -> dict:
            """Add task instructions to the question."""
            if not isinstance(input, dict) or "question" not in input:
                raise ValueError(
                    f"Expected input to be a dict with key `question`. "
                    f"Found {type(input)}."
                )

            input = input.copy()
            input["question"] = (
                f"{self.task.instructions}\nWrite down your answer, "
                f"but do not explain it. Input: `{input['question']}`"
            )
            return input

        agent = create_agent(env.tools)  # type: ignore
        # Returns `state` in the output if the environment has a state reader
        # makes sure that `output` is always in the output

        if kwargs:
            agent = agent.bind(**kwargs)

        runnable = _add_task_instructions | apply_agent_executor_adapter(
            agent, state_reader=env.read_state
        )

        if self.rate_limiter:  # Add a rate limiter
            runnable = rate_limiting.with_rate_limit(runnable, self.rate_limiter)

        return runnable



from langchain_benchmarks.rag.tasks.langchain_docs.task import LANGCHAIN_DOCS_TASK
from langchain_benchmarks.rag.tasks.multi_modal_slide_decks.task import (
    MULTI_MODAL_SLIDE_DECKS_TASK,
)
from langchain_benchmarks.rag.tasks.semi_structured_reports.task import (
    SEMI_STRUCTURED_REPORTS_TASK,
)

# Please keep this sorted
__all__ = [
    "LANGCHAIN_DOCS_TASK",
    "SEMI_STRUCTURED_REPORTS_TASK",
    "MULTI_MODAL_SLIDE_DECKS_TASK",
]



"""Module contains standard evaluators for agents.

Requirements:

* Agents must output "intermediate_steps" in their run outputs.
* The dataset must have "expected_steps" in its outputs.
"""
import re
from typing import Any, Literal, Optional, Union

from langchain.callbacks.manager import collect_runs
from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.evaluation import EvaluatorType, StringEvaluator, load_evaluator
from langchain.evaluation.schema import StringEvaluator
from langchain.smith import RunEvalConfig
from langchain_core.language_models import BaseChatModel, BaseLanguageModel
from langsmith.evaluation.evaluator import (
    EvaluationResult,
    EvaluationResults,
    RunEvaluator,
)
from langsmith.schemas import Example, Run

from langchain_benchmarks.tool_usage.prompts import (
    QA_TEMPLATE_FOR_MULTIVERSE_MATH,
    QA_TEMPLATE_FOR_MULTIVERSE_MATH_WITHOUT_QUESTION,
)

OutputEvaluation = Literal["qa", "qa_math", "none", "qa_math_without_question"]


class QAMathEvaluator(StringEvaluator):
    """An LLM-based relevance evaluator."""

    def __init__(self, chat_model: BaseChatModel) -> None:
        """Initialize the evaluator."""
        self.eval_chain = QA_TEMPLATE_FOR_MULTIVERSE_MATH_WITHOUT_QUESTION | chat_model

    @property
    def evaluation_name(self) -> str:
        """Return the name of the evaluator."""
        return "QAMathEvaluator"

    @property
    def requires_reference(self) -> bool:
        return True

    @property
    def requires_input(self) -> bool:
        return False

    def _evaluate_strings(
        self,
        prediction: str,
        input: Optional[str] = None,
        reference: Optional[str] = None,
        **kwargs: Any,
    ) -> dict:
        """Evaluate the prediction against the reference."""
        result = self.eval_chain.invoke(
            {"answer": reference, "result": prediction}, **kwargs
        )
        if result.content.startswith("CORRECT"):
            return {"score": 1}
        else:
            return {"score": 0}


def compare_outputs(
    run_outputs: dict,
    example_outputs: dict,
    run_inputs: dict,
    *,
    qa_evaluator: Optional[StringEvaluator] = None,
) -> EvaluationResults:
    """Compare the outputs of a run to the expected outputs."""
    if "intermediate_steps" in run_outputs:
        intermediate_steps = run_outputs["intermediate_steps"]
        # Since we are comparing to the tool names, we now need to get that
        # Intermediate steps is a Tuple[AgentAction, Any]
        # The first element is the action taken
        # The second element is the observation from taking that action
        actual_steps = [action.tool for action, _ in intermediate_steps]
    elif "actual_steps" in run_outputs:
        actual_steps = run_outputs["actual_steps"]
    else:
        raise ValueError(
            "Please make sure that your agent outputs 'intermediate_steps or "
            "'actual_steps'"
        )

    expected_steps = example_outputs["expected_steps"]

    order_matters = example_outputs.get("order_matters", True)

    if order_matters:
        # If the order matters trajectory must be the same as expected trajectory
        trajectory_score = int(actual_steps == expected_steps)
    else:
        # If order does not matter, then we compare the trajectories after sorting
        # them. This will make sure that the number of times each tool is used
        # is the same, but the order does not matter.
        trajectory_score = int(sorted(actual_steps) == sorted(expected_steps))

    # Just score it based on whether it is correct or not
    step_fraction = len(actual_steps) / len(expected_steps)

    # Add trajectory results
    results = [
        EvaluationResult(
            key="Intermediate steps correctness",
            score=trajectory_score,
            comment=f"Order matters={order_matters}",
        ),
        EvaluationResult(
            key="# steps / # expected steps",
            score=step_fraction,
        ),
    ]

    # Evaluate state score
    # This will need to be evolved it's too simple.
    if "state" in run_outputs and "state" in example_outputs:
        state = run_outputs["state"]
        example_state = example_outputs["state"]
        results.append(
            EvaluationResult(
                key="Correct Final State",
                score=int(state == example_state),
            )
        )

    if "output" in run_outputs and qa_evaluator:
        output = run_outputs["output"]
        with collect_runs() as cb:
            if isinstance(qa_evaluator, QAMathEvaluator):
                qa_results = qa_evaluator.evaluate_strings(
                    prediction=output,
                    reference=example_outputs["reference"],
                )
            else:
                qa_results = qa_evaluator.evaluate_strings(
                    prediction=output,
                    reference=example_outputs["reference"],
                    input=run_inputs["question"],
                )
        results.append(
            EvaluationResult(
                key="correctness",
                score=qa_results["score"],
                source_run_id=cb.traced_runs[0].id,
            )
        )

    return {"results": results}


class AgentTrajectoryEvaluator(RunEvaluator):
    """An evaluator that can be used in conjunction with a standard agent interface."""

    def __init__(
        self,
        eval_llm: Union[BaseLanguageModel, BaseChatModel, None] = None,
        output_evaluation: Literal["qa", "none", "qa_math"] = "qa",
    ) -> None:
        """Initialize the evaluator."""
        if output_evaluation == "none":
            if eval_llm is not None:
                raise ValueError(
                    "If output_evaluation is 'none', then eval_llm must be None"
                )
            qa_evaluator = None
        else:
            eval_llm = eval_llm or ChatOpenAI(
                model="gpt-4",
                temperature=0,
                model_kwargs={"seed": 42},
                max_retries=1,
                request_timeout=60,
            )
            if output_evaluation == "qa":
                qa_evaluator = load_evaluator(EvaluatorType.QA, llm=eval_llm)
            elif output_evaluation == "qa_math":
                qa_evaluator = load_evaluator(
                    EvaluatorType.QA,
                    llm=eval_llm,
                    prompt=QA_TEMPLATE_FOR_MULTIVERSE_MATH,
                )
            elif output_evaluation == "qa_math_without_question":
                qa_evaluator = QAMathEvaluator(eval_llm)
            else:
                raise ValueError(
                    f"output_evaluation must be one of 'qa' or 'none', "
                    f"got {output_evaluation}"
                )

        self.qa_evaluator = qa_evaluator
        self.output_evaluation = output_evaluation

    def evaluate_run(
        self, run: Run, example: Optional[Example] = None
    ) -> EvaluationResults:
        # The run is the run from the agent
        if run.outputs is None:
            raise ValueError("Run outputs cannot be None")

        # The example is the example from the dataset
        if example is None:
            raise ValueError("Example cannot be None")

        if (
            "intermediate_steps" not in run.outputs
            and "actual_steps" not in run.outputs
        ):
            raise ValueError(
                "Please make sure that your agent outputs 'intermediate_steps or "
                "'actual_steps'"
            )

        if "expected_steps" not in example.outputs:
            raise ValueError(
                "Please make sure that your dataset contains 'expected_steps'"
            )

        return compare_outputs(
            run.outputs,
            example.outputs,
            qa_evaluator=self.qa_evaluator,
            run_inputs=run.inputs,
        )


def get_eval_config(
    *,
    eval_llm: Union[BaseLanguageModel, BaseChatModel, None] = None,
    output_evaluation: OutputEvaluation = "qa",
) -> RunEvalConfig:
    """Get the default evaluator for the environment.

    Args:
        eval_llm: The language model to use for grading the `output` response
        output_evaluation: how to evaluate the output of the agent.
            - 'qa' will use the qa evaluator to compare the output to the reference.
            - 'qa_math' will use the qa evaluator to compare the output to the reference
               using a prompt that better works for multiverse math.
            - 'none' will not evaluate the output of the agent -- in some cases
              it's only relevant to evaluate how the agent used tools, not what
              its output.

    Returns:
        A RunEvalConfig that can be used to evaluate the environment
    """
    return RunEvalConfig(
        custom_evaluators=[
            AgentTrajectoryEvaluator(
                eval_llm=eval_llm, output_evaluation=output_evaluation
            )
        ]
    )



from langchain.prompts import ChatPromptTemplate

from langchain_benchmarks.extraction.tasks.chat_extraction.evaluators import (
    get_eval_config,
)
from langchain_benchmarks.extraction.tasks.chat_extraction.schema import GenerateTicket
from langchain_benchmarks.schema import ExtractionTask

# This is a default prompt that works reasonably for OpenAI models.

DEFAULT_CHAT_MODEL_PROMPT = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are a helpdesk assistant responsible with extracting information"
            " and generating tickets. Dialogues are between a user and"
            " a support engineer.",
        ),
        (
            "user",
            "Generate a ticket for the following question-response pair:\n"
            "<Dialogue>\n{dialogue}\n</Dialogue>",
        ),
    ]
)


CHAT_EXTRACTION_TASK = ExtractionTask(
    name="Chat Extraction",
    dataset_id="https://smith.langchain.com/public/00f4444c-9460-4a82-b87a-f50096f1cfef/d",
    schema=GenerateTicket,
    description="""A dataset meant to test the ability of an LLM to extract and infer
structured information from a dialogue. The dialogue is between a user and a support
engineer. Outputs should be structured as a JSON object and test both the ability
of the LLM to correctly structure the information and its ability to perform simple 
classification tasks.""",
    instructions=DEFAULT_CHAT_MODEL_PROMPT,
)


__all__ = ["CHAT_EXTRACTION_TASK", "get_eval_config"]



"""Implementation of a rate limiter based on a token bucket."""
import threading
import time
from typing import Any, Optional

from langchain.schema.runnable import Runnable, RunnableLambda
from langchain.schema.runnable.utils import Input, Output


class RateLimiter:
    def __init__(
        self,
        *,
        requests_per_second: float = 1,
        check_every_n_seconds: float = 0.1,
        max_bucket_size: float = 1,
    ) -> None:
        """A rate limiter based on a token bucket.

        These *tokens* have NOTHING to do with LLM tokens. They are just
        a way to keep track of how many requests can be made at a given time.

        This rate limiter is designed to work in a threaded environment.

        It works by filling up a bucket with tokens at a given rate. Each
        request consumes a given number of tokens. If there are not enough
        tokens in the bucket, the request is blocked until there are enough
        tokens.

        Args:
            requests_per_second: The number of tokens to add per second to the bucket.
                Must be at least 1. The tokens represent "credit" that can be used
                to make requests.
            check_every_n_seconds: check whether the tokens are available
                every this many seconds. Can be a float to represent
                fractions of a second.
            max_bucket_size: The maximum number of tokens that can be in the bucket.
                This is used to prevent bursts of requests.
        """
        # Number of requests that we can make per second.
        self.requests_per_second = requests_per_second
        # Number of tokens in the bucket.
        self.available_tokens = 0.0
        self.max_bucket_size = max_bucket_size
        # A lock to ensure that tokens can only be consumed by one thread
        # at a given time.
        self._consume_lock = threading.Lock()
        # The last time we tried to consume tokens.
        self.last: Optional[time.time] = None
        self.check_every_n_seconds = check_every_n_seconds

    def _consume(self) -> bool:
        """Consume the given amount of tokens if possible.

        Returns:
            True means that the tokens were consumed, and the caller can proceed to
            make the request. A False means that the tokens were not consumed, and
            the caller should try again later.
        """
        with self._consume_lock:
            now = time.time()

            # initialize on first call to avoid a burst
            if self.last is None:
                self.last = now

            elapsed = now - self.last

            if elapsed * self.requests_per_second >= 1:
                self.available_tokens += elapsed * self.requests_per_second
                self.last = now

            # Make sure that we don't exceed the bucket size.
            # This is used to prevent bursts of requests.
            self.available_tokens = min(self.available_tokens, self.max_bucket_size)

            # As long as we have at least one token, we can proceed.
            if self.available_tokens >= 1:
                self.available_tokens -= 1
                return True

            return False

    def wait(self) -> None:
        """Blocking call to wait until the given number of tokens are available."""
        while not self._consume():
            time.sleep(self.check_every_n_seconds)


def with_rate_limit(
    runnable: Runnable[Input, Output],
    rate_limiter: RateLimiter,
) -> Runnable[Input, Output]:
    """Add a rate limiter to the runnable.

    Args:
        runnable: The runnable to throttle.
        rate_limiter: The throttle to use.

    Returns:
        A runnable lambda that acts as a throttled passthrough.
    """

    def _wait(input: dict, **kwargs: Any) -> dict:
        """Wait for the rate limiter to allow the request to proceed."""
        rate_limiter.wait()
        return input

    return RunnableLambda(_wait).with_config({"name": "Wait"}) | runnable



import logging
import os
import zipfile
from pathlib import Path
from typing import Iterable, Optional

from langchain_benchmarks.rag.utils._downloading import (
    fetch_remote_file,
    is_folder_populated,
)

logger = logging.getLogger(__name__)
_DIRECTORY = Path(os.path.abspath(__file__)).parent
# Stores the zipped pdfs for this dataset
REMOTE_DOCS_FILE = "https://storage.googleapis.com/benchmarks-artifacts/langchain-docs-benchmarking/multi_modal_slide_decks.zip"
DOCS_DIR = _DIRECTORY / "pdfs"


def fetch_raw_docs(
    filename: Optional[str] = None, docs_dir: Optional[str] = None
) -> None:
    filename = filename or _DIRECTORY / Path(REMOTE_DOCS_FILE).name
    docs_dir = docs_dir or DOCS_DIR
    if not is_folder_populated(docs_dir):
        fetch_remote_file(REMOTE_DOCS_FILE, filename)
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(docs_dir)

        os.remove(filename)


def get_file_names() -> Iterable[Path]:
    fetch_raw_docs()
    # Traverse the directory and partition the pdfs
    for path in DOCS_DIR.rglob("*.pdf"):
        # Ignore __MACOSX
        if "__MACOSX" in str(path):
            continue
        yield path



"""Schema for the Langchain Benchmarks."""
from __future__ import annotations

import dataclasses
import importlib
import urllib
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Type, Union

from langchain.prompts import ChatPromptTemplate
from langchain.schema import BaseRetriever
from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.smith import RunEvalConfig
from langchain.tools import BaseTool
from langchain_core.language_models import BaseChatModel, BaseLanguageModel
from pydantic import BaseModel
from tabulate import tabulate
from typing_extensions import Literal


@dataclasses.dataclass(frozen=True)
class ToolUsageEnvironment:
    """An instance of an environment for tool usage."""

    tools: List[BaseTool]
    """The tools that can be used in the environment."""

    read_state: Optional[Callable[[], Any]] = None
    """A function that returns the current state of the environment."""


@dataclasses.dataclass(frozen=True)
class BaseTask:
    """A definition of a task."""

    name: str
    """The name of the environment."""

    dataset_id: str
    """The ID of the langsmith public dataset.

    This dataset contains expected inputs/outputs for the environment, and
    can be used to evaluate the performance of a model/agent etc.
    """

    description: str
    """Description of the task for a data science practitioner.

    This can contain information about the task, the dataset, the tools available
    etc.
    """

    @property
    def _dataset_link(self) -> str:
        """Return a link to the dataset."""
        dataset_url = (
            self.dataset_id
            if self.dataset_id.startswith("http")
            else f"https://smith.langchain.com/public/{self.dataset_id}/d"
        )
        parsed_url = urllib.parse.urlparse(dataset_url)
        # Extract the UUID from the path
        path_parts = parsed_url.path.split("/")
        token_uuid = path_parts[-2] if len(path_parts) >= 2 else "Link"
        return (
            f'<a href="{dataset_url}" target="_blank" rel="noopener">{token_uuid}</a>'
        )

    @property
    def _table(self) -> List[List[str]]:
        """Return a table representation of the environment."""
        return [
            ["Name", self.name],
            ["Type", self.type],
            ["Dataset ID", self._dataset_link],
            ["Description", self.description],
        ]

    def _repr_html_(self) -> str:
        """Return an HTML representation of the environment."""
        return tabulate(
            self._table,
            tablefmt="unsafehtml",
        )

    @property
    def type(self) -> str:
        """Return the type of the task."""
        return self.__class__.__name__


@dataclasses.dataclass(frozen=True)
class ToolUsageTask(BaseTask):
    """A definition for a task."""

    create_environment: Callable[[], ToolUsageEnvironment]
    """Factory that returns an environment."""

    instructions: str
    """Instructions for the agent/chain/llm."""

    eval_params: Dict[str, Any]
    """Used to parameterize differences in the evaluation of the task.
    
    These are passed to the standard factory method for creating an evaluator
    for tool usage.
    
    An example, for MultiVerse math the `output_evaluation` parameter is set to
    `qa_math` to use a different prompt for evaluating the output of the agent.
    
    This prompt performs better at comparing the output of the agent against
    the reference output.
    """

    def get_eval_config(self, **params: Any) -> RunEvalConfig:
        """Get the default evaluator for the environment."""
        # Import locally to avoid potential circular imports in the future.
        from langchain_benchmarks.tool_usage.evaluators import get_eval_config

        finalized_params = {**self.eval_params, **params}
        return get_eval_config(**finalized_params)


@dataclasses.dataclass(frozen=True)
class ExtractionTask(BaseTask):
    """A definition for an extraction task."""

    schema: Type[BaseModel]
    """Get schema that specifies what should be extracted."""

    # We might want to make this optional / or support more types
    # and add validation, but let's wait until we have more examples
    instructions: ChatPromptTemplate
    """Get the prompt for the task.
    
    This is the default prompt to use for the task.
    """


@dataclasses.dataclass(frozen=True)
class RetrievalTask(BaseTask):
    get_docs: Optional[Callable[..., Iterable[Document]]] = None
    """A function that returns the documents to be indexed."""
    retriever_factories: Dict[
        str, Callable[[Embeddings], BaseRetriever]
    ] = dataclasses.field(default_factory=dict)  # noqa: F821
    """Factories that index the docs using the specified strategy."""
    architecture_factories: Dict[
        str, Callable[[Embeddings], BaseRetriever]
    ] = dataclasses.field(default_factory=dict)  # noqa: F821
    """Factories methods that help build some off-the-shelf architectures。"""

    @property
    def _table(self) -> List[List[str]]:
        """Get information about the task."""
        table = super()._table
        return table + [
            ["Retriever Factories", ", ".join(self.retriever_factories.keys())],
            ["Architecture Factories", ", ".join(self.architecture_factories.keys())],
            ["get_docs", self.get_docs],
        ]


@dataclasses.dataclass(frozen=False)
class Registry:
    tasks: List[BaseTask]

    def get_task(self, name_or_id: Union[int, str]) -> BaseTask:
        """Get the task with the given name or ID."""
        if isinstance(name_or_id, int):
            return self.tasks[name_or_id]

        for env in self.tasks:
            if env.name == name_or_id:
                return env
        raise ValueError(f"Unknown task {name_or_id}")

    def __post_init__(self) -> None:
        """Validate that all the tasks have unique names and IDs."""
        seen_names = set()
        for task in self.tasks:
            if task.name in seen_names:
                raise ValueError(
                    f"Duplicate task name {task.name}. " f"Task names must be unique."
                )
            seen_names.add(task.name)

    def __len__(self) -> int:
        """Return the number of tasks in the registry."""
        return len(self.tasks)

    def __iter__(self) -> Iterable[BaseTask]:
        """Iterate over the tasks in the registry."""
        return iter(self.tasks)

    def _repr_html_(self) -> str:
        """Return an HTML representation of the registry."""
        headers = [
            "Name",
            "Type",
            "Dataset ID",
            "Description",
        ]
        table = [
            [
                task.name,
                task.__class__.__name__,
                task._dataset_link,
                task.description,
            ]
            for task in self.tasks
        ]
        return tabulate(table, headers=headers, tablefmt="unsafehtml")

    def filter(
        self,
        Type: Optional[str],
        dataset_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Registry:
        """Filter the tasks in the registry."""
        tasks = self.tasks
        if Type is not None:
            tasks = [task for task in tasks if task.__class__.__name__ == Type]
        if dataset_id is not None:
            tasks = [task for task in tasks if task.dataset_id == dataset_id]
        if name is not None:
            tasks = [task for task in tasks if task.name == name]
        if description is not None:
            tasks = [
                task
                for task in tasks
                if description.lower() in task.description.lower()
            ]
        return Registry(tasks=tasks)

    def __getitem__(self, key: Union[int, str, slice]) -> Union[BaseTask, Registry]:
        """Get an environment from the registry."""
        if isinstance(key, slice):
            return Registry(tasks=self.tasks[key])
        elif isinstance(key, (int, str)):
            # If key is an integer, return the corresponding environment
            return self.get_task(key)
        else:
            raise TypeError("Key must be an integer or a slice.")

    def add(self, task: BaseTask) -> None:
        if not isinstance(task, BaseTask):
            raise TypeError("Only tasks can be added to the registry.")
        self.tasks.append(task)


Provider = Literal["fireworks", "openai", "anthropic", "anyscale"]
ModelType = Literal["chat", "llm"]
AUTHORIZED_NAMESPACES = {"langchain", "langchain_google_genai"}


def _get_model_class_from_path(
    path: str,
) -> Union[Type[BaseChatModel], Type[BaseLanguageModel]]:
    """Get the class of the model."""
    module_name, attribute_name = path.rsplit(".", 1)
    top_namespace = path.split(".")[0]

    if top_namespace not in AUTHORIZED_NAMESPACES:
        raise ValueError(
            f"Unauthorized namespace {top_namespace}. "
            f"Authorized namespaces are: {AUTHORIZED_NAMESPACES}"
        )

    # Import the module dynamically
    module = importlib.import_module(module_name)
    model_class = getattr(module, attribute_name)
    if not issubclass(model_class, (BaseLanguageModel, BaseChatModel)):
        raise ValueError(
            f"Model class {model_class} is not a subclass of BaseLanguageModel"
        )
    return model_class


def _get_default_path(provider: str, type_: ModelType) -> str:
    """Get the default path for a model."""
    paths = {
        ("fireworks", "chat"): "langchain.chat_models.fireworks.ChatFireworks",
        ("fireworks", "llm"): "langchain.llms.fireworks.Fireworks",
        ("anyscale", "chat"): "langchain.chat_models.anyscale.ChatAnyscale",
        ("anyscale", "llm"): "langchain.llms.anyscale.Anyscale",
        ("openai", "chat"): "langchain.chat_models.openai.ChatOpenAI",
        ("openai", "llm"): "langchain.llms.openai.OpenAI",
        ("anthropic", "chat"): "langchain.chat_models.anthropic.ChatAnthropic",
        (
            "google-genai",
            "chat",
        ): "langchain_google_genai.chat_models.ChatGoogleGenerativeAI",
    }

    if (provider, type_) not in paths:
        raise ValueError(f"Unknown provider {provider} and type {type_}")

    return paths[(provider, type_)]


def _get_default_url(provider: str, type_: ModelType) -> Optional[str]:
    """Get default URL to API page for model."""
    if provider == "fireworks":
        return "https://app.fireworks.ai/models"
    elif provider == "openai":
        return "https://platform.openai.com/docs/models"
    elif provider == "anthropic":
        return "https://docs.anthropic.com/claude/reference/selecting-a-model"
    elif provider == "anyscale":
        return "https://docs.endpoints.anyscale.com/category/supported-models"
    elif provider == "google-genai":
        return "https://ai.google.dev/"
    else:
        return None


@dataclasses.dataclass(frozen=True)
class RegisteredModel:
    """Descriptive information about a model.

    This information can be used to instantiate the underlying model.
    """

    name: str
    provider: Provider
    description: str
    params: Dict[str, Any]
    type: ModelType
    # Path to the model class.
    # For example, "langchain.chat_models.anthropic import ChatAnthropicModel"
    path: Optional[str] = None  # If not provided, will use default path
    url: Optional[str] = None  # If not provided, will use default URL

    def get_model(
        self, *, model_params: Optional[Dict[str, Any]] = None
    ) -> Union[BaseChatModel, BaseLanguageModel]:
        """Get the class of the model."""
        all_params = {**self.params, **(model_params or {})}
        model_class = _get_model_class_from_path(self.model_path)
        return model_class(**all_params)

    @property
    def model_path(self) -> str:
        """Get the path of the model."""
        return self.path or _get_default_path(self.provider, self.type)

    @property
    def model_url(self) -> Optional[str]:
        """Get the URL of the model."""
        return self.url or _get_default_url(self.provider, self.type)

    @property
    def _table(self) -> List[List[str]]:
        """Return a table representation of the environment."""
        if self.model_path:
            url = (
                f'<a href="{self.model_path}" target="_blank" rel="noopener">'
                "ModelPage"
                "</a>"
            )
        else:
            url = ""
        return [
            ["name", self.name],
            ["type", self.type],
            ["provider", self.provider],
            ["description", self.description],
            ["model_path", self.model_path],
            ["url", url],
        ]

    def _repr_html_(self) -> str:
        """Return an HTML representation of the environment."""
        return tabulate(
            self._table,
            tablefmt="unsafehtml",
        )


StrFilter = Union[None, str, Sequence[str]]


def _is_in_filter(actual_value: str, filter_value: StrFilter) -> bool:
    """Filter for a string attribute."""
    if filter_value is None:
        return True

    if isinstance(filter_value, str):
        return actual_value == filter_value

    return actual_value in filter_value


@dataclasses.dataclass(frozen=False)
class ModelRegistry:
    registered_models: Sequence[RegisteredModel]

    def __post_init__(self) -> None:
        """Validate that all the tasks have unique names and IDs."""
        seen_names = set()
        for model in self.registered_models:
            if model.name in seen_names:
                raise ValueError(
                    f"Duplicate model name {model.name}. " f"Task names must be unique."
                )
            seen_names.add(model.name)

    def get_model(self, name: str) -> Optional[RegisteredModel]:
        """Get model info."""
        return next(model for model in self.registered_models if model.name == name)

    def filter(
        self,
        *,
        type: StrFilter = None,
        name: StrFilter = None,
        provider: StrFilter = None,
    ) -> ModelRegistry:
        """Filter the tasks in the registry."""
        models = self.registered_models
        selected_models = []

        for model in models:
            if not _is_in_filter(model.type, type):
                continue
            if not _is_in_filter(model.name, name):
                continue
            if not _is_in_filter(model.provider, provider):
                continue
            selected_models.append(model)
        return ModelRegistry(registered_models=selected_models)

    def _repr_html_(self) -> str:
        """Return an HTML representation of the registry."""
        headers = [
            "Name",
            "Type",
            "Provider",
            "Description",
        ]
        table = [
            [
                model.name,
                model.type,
                model.provider,
                model.description,
            ]
            for model in self.registered_models
        ]
        return tabulate(table, headers=headers, tablefmt="unsafehtml")

    def __len__(self) -> int:
        """Return the number of tasks in the registry."""
        return len(self.registered_models)

    def __contains__(self, item: Any) -> bool:
        """Return whether the registry contains the given model."""
        return self.get_model(item) is not None

    def __iter__(self) -> Iterable[RegisteredModel]:
        """Iterate over the tasks in the registry."""
        return iter(self.registered_models)

    def __getitem__(
        self, key: Union[int, str, slice]
    ) -> Union[RegisteredModel, ModelRegistry]:
        """Get an environment from the registry."""
        if isinstance(key, slice):
            return ModelRegistry(registered_models=self.registered_models[key])
        elif isinstance(key, (int, str)):
            # If key is an integer, return the corresponding environment
            if isinstance(key, str):
                return self.get_model(key)
            else:
                return self.registered_models[key]
        else:
            raise TypeError("Key must be an integer or a slice.")



from typing import Optional

from langchain.base_language import BaseLanguageModel
from langchain.chat_models import ChatOpenAI
from langchain.schema.retriever import BaseRetriever
from langchain.schema.runnable import Runnable

from langchain_benchmarks.rag.tasks.langchain_docs.architectures.crqa import (
    create_response_chain,
    get_default_response_generator,
)


def default_response_chain(
    retriever: BaseRetriever,
    response_generator: Optional[Runnable] = None,
    llm: Optional[BaseLanguageModel] = None,
) -> None:
    """Get the chain responsible for generating a response based on the retrieved documents."""
    response_generator = response_generator or get_default_response_generator(
        llm=llm or ChatOpenAI(model="gpt-3.5-turbo-16k", model_kwargs={"seed": 42})
    )
    return create_response_chain(
        response_generator=response_generator, retriever=retriever
    )


ARCH_FACTORIES = {
    "conversational-retrieval-qa": default_response_chain,
}



"""Prototyping code for rendering function definitions, invocations, and results.

Types are simplified for now to `str`.

We should actually support something like pydantic or jsonschema for the types, so
we can expand them recursively for nested types.
"""
import abc
from typing import Any, List, Optional

from typing_extensions import NotRequired, TypedDict


class Parameter(TypedDict):
    """Representation for a parameter."""

    name: str
    type: str
    description: str


class Arguments(TypedDict):
    """Arguments are passed to a function during function invocation."""

    name: Optional[str]
    value: Any


class ReturnValue(TypedDict):
    """Representation for a return value of a function call."""

    type: str
    description: NotRequired[str]


class FunctionDefinition(TypedDict):
    """Representation for a function."""

    name: str
    description: str  # Function description
    parameters: List[Parameter]
    return_value: ReturnValue


class FunctionInvocation(TypedDict):
    """Representation for a function invocation."""

    id: NotRequired[str]
    name: str
    arguments: List[Arguments]


class FunctionResult(TypedDict):
    """Representation for a function result."""

    id: NotRequired[str]
    name: str
    result: Optional[str]
    error: Optional[str]


class Visitor(abc.ABC):
    @abc.abstractmethod
    def visit_function_definition(self, function_definition: FunctionDefinition) -> str:
        """Render a function."""

    @abc.abstractmethod
    def visit_function_definitions(
        self, function_definitions: List[FunctionDefinition]
    ) -> str:
        """Render a function."""

    @abc.abstractmethod
    def visit_function_invocation(self, function_invocation: FunctionInvocation) -> str:
        """Render a function invocation."""

    @abc.abstractmethod
    def visit_function_result(self, function_result: FunctionResult) -> str:
        """Render a function result."""


class AstPrinter(Visitor):
    """Print the AST."""


class XMLEncoder(AstPrinter):
    def visit_function_definition(self, function_definition: FunctionDefinition) -> str:
        """Render a function."""
        parameters_lines = []

        for parameter in function_definition["parameters"]:
            parameters_lines.extend(
                [
                    "<parameter>",
                    f"<name>{parameter['name']}</name>",
                    f"<type>{parameter['type']}</type>",
                    f"<description>{parameter['description']}</description>",
                    "</parameter>",
                ]
            )
        lines = [
            "<function>",
            f"<function_name>{function_definition['name']}</function_name>",
            "<description>",
            f"{function_definition['description']}",
            "</description>",
            "<parameters>",
            *parameters_lines,
            "</parameters>",
            "<return_value>",
            f"<type>{function_definition['return_value']['type']}</type>",
        ]
        if function_definition["return_value"].get("description"):
            lines.append(
                f"<description>{function_definition['return_value']['description']}"
                f"</description>"
            )

        lines.extend(["</return_value>", "</function>"])
        return "\n".join(lines)

    def visit_function_definitions(
        self, function_definitions: List[FunctionDefinition]
    ) -> str:
        """Render a function."""
        strs = [
            self.visit_function_definition(function_definition)
            for function_definition in function_definitions
        ]
        return "<functions>\n" + "\n".join(strs) + "\n</functions>"

    def visit_function_invocation(self, invocation: FunctionInvocation) -> str:
        """Render a function invocation."""
        arguments_as_strings = [
            "<argument>\n"
            f"<name>{argument['name']}</name>\n"
            f"<value>{argument['value']}</value>\n"
            "</argument>\n"
            for argument in invocation["arguments"]
        ]
        lines = ["<function_invocation>"]

        if invocation.get("id"):
            lines.append(f"<id>{invocation['id']}</id>")

        lines.extend(
            [
                f"<function_name>{invocation['name']}</function_name>\n"
                "<arguments>\n"
                f"{''.join(arguments_as_strings)}"  # Already includes trailing newline
                "</arguments>\n"
                "</function_invocation>"
            ]
        )
        return "\n".join(lines)

    def visit_function_result(self, function_result: FunctionResult) -> str:
        """Render a function result."""
        lines = [
            "<function_result>",
        ]

        if function_result.get("id"):
            lines.append(f"<id>{function_result['id']}</id>")

        lines.append(f"<function_name>{function_result['name']}</function_name>")

        if function_result["error"]:
            lines.extend(
                [
                    f"<error>{function_result['error']}</error>",
                ]
            )
        else:
            lines.append(
                f"<result>{function_result['result']}</result>",
            )

        lines.append("</function_result>")

        return "\n".join(lines)


class TypeScriptEncoder(AstPrinter):
    def visit_function_definition(self, function_definition: FunctionDefinition) -> str:
        """Render a function."""
        parameters_as_strings = [
            f"{parameter['name']}: {parameter['type']}"
            for parameter in function_definition["parameters"]
        ]
        # Let's use JSdoc style comments
        # First the function description
        lines = [
            f"// {function_definition['description']}",
            # Then the parameter descriptions
            *[
                f"// @param {parameter['name']} {parameter['description']}"
                for parameter in function_definition["parameters"]
            ],
            # Then the return value description
            f"// @returns {function_definition['return_value']['description']}",
            # Then the function definition
            f"function {function_definition['name']}("
            f"{', '.join(parameters_as_strings)}): "
            f"{function_definition['return_value']['type']};",
        ]

        # finally join
        function = "\n".join(lines)
        return function

    def visit_function_definitions(
        self, function_definitions: List[FunctionDefinition]
    ) -> str:
        """Render a function."""
        strs = [
            self.visit_function_definition(function_definition)
            for function_definition in function_definitions
        ]
        return "\n\n".join(strs)

    def visit_function_invocation(self, invocation: FunctionInvocation) -> str:
        """Render a function invocation."""
        arguments_as_strings = [
            f"{argument['name']}: {argument['value']}"
            for argument in invocation["arguments"]
        ]
        lines = [f"{invocation['name']}(" f"{', '.join(arguments_as_strings)});"]
        return "\n".join(lines)

    def visit_function_result(self, function_result: FunctionResult) -> str:
        """Render a function result."""
        lines = []
        if function_result["error"]:
            lines.append(f"ERROR: {function_result['error']}")
        else:
            lines.append(f"> {function_result['result']}")
        if function_result.get("id"):
            lines.append(f"// ID: {function_result['id']}")
        return "\n".join(lines)



from langchain_benchmarks.extraction.evaluators import get_eval_config
from langchain_benchmarks.extraction.implementations import (
    create_openai_function_based_extractor,
)

# Keep this sorted
__all__ = [
    "get_eval_config",
    "create_openai_function_based_extractor",
]



"""Code for creating an agent factory for evaluating tool usage tasks."""
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

from langchain.agents import AgentExecutor
from langchain.agents.format_scratchpad.openai_tools import (
    format_to_openai_tool_messages,
)
from langchain.agents.output_parsers.openai_tools import OpenAIToolsAgentOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema.runnable import Runnable
from langchain.tools.render import format_tool_to_openai_tool
from langchain_core.language_models import BaseChatModel, BaseLanguageModel
from langchain_core.language_models.base import LanguageModelInput
from langchain_core.messages import BaseMessage
from langchain_core.pydantic_v1 import BaseModel

from langchain_benchmarks import model_registry, rate_limiting
from langchain_benchmarks.model_registration import RegisteredModel
from langchain_benchmarks.schema import ToolUsageTask
from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter

# PUBLIC API


def _bind_tools(
    llm: BaseChatModel,
    tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable]],
    tool_choice: Optional[str] = None,
    json_mode: bool = False,
    **kwargs: Any,
) -> Runnable[LanguageModelInput, BaseMessage]:
    """Bind tools (and other objects) to this chat model.

    Args:
        tools: A list of tool definitions to bind to this chat model.
            Can be  a dictionary, pydantic model, or callable. Pydantic
            models and callables will be automatically converted to
            their schema dictionary representation.
        tool_choice: Which tool to require the model to call.
            Must be the name of the single provided tool or
            "auto" to automatically determine which tool to call
            (if any).
        json_mode: Whether to set JSON mode for the tool call.
            This guarantees the model will respond in valid JSON
            (unless truncated).
        kwargs: Any additional parameters to pass to the
            :class:`~langchain.runnable.Runnable` constructor.

    """
    formatted_tools: List[Dict[str, Union[str, dict]]] = [
        format_tool_to_openai_tool(tool) for tool in tools
    ]
    if tool_choice is not None:
        if not formatted_tools:
            raise ValueError(
                "When specifying `tool_choice`, you must provide at least one " "tool."
            )
        tool_names = [tool["function"]["name"] for tool in formatted_tools]
        if not any(tool_name == tool_choice for tool_name in tool_names):
            raise ValueError(
                f"Tool choice {tool_choice} was specified, but the only "
                f"provided tools were {tool_names}."
            )
        tool_choice_ = {"type": "function", "function": {"name": tool_choice}}
        kwargs = {**kwargs, "tool_choice": tool_choice_}
    if json_mode:
        kwargs = {**kwargs, "response_format": {"type": "json_object"}}
    return llm.bind(
        tools=formatted_tools,
        **kwargs,
    )


class OpenAIAgentFactory:
    def __init__(
        self,
        task: ToolUsageTask,
        *,
        model: Union[
            str, RegisteredModel, BaseLanguageModel, BaseChatModel
        ] = "gpt-3.5-turbo-16k",
        rate_limiter: Optional[rate_limiting.RateLimiter] = None,
        num_retries: int = 0,
    ) -> None:
        """Create an OpenAI agent factory for the given task.

        Args:
            task: The task to create an agent factory for.
            model: The model to use -- this must be an open AI model.
            rate_limiter: The rate limiter to use
        """
        self.task = task
        self.model = model
        self.rate_limiter = rate_limiter
        self.num_retries = num_retries

    def _create_model(self) -> Union[BaseChatModel, BaseLanguageModel]:
        if isinstance(self.model, RegisteredModel):
            return self.model.get_model(
                model_params={"temperature": 0, "model_kwargs": {"seed": 0}}
            )
        elif isinstance(self.model, (BaseChatModel, BaseLanguageModel)):
            return self.model
        elif isinstance(self.model, str):
            if self.model in model_registry:
                registered_model = model_registry.get_model(self.model)
                model_instance = registered_model.get_model(
                    model_params={"temperature": 0, "model_kwargs": {"seed": 0}}
                )
                return model_instance
            else:
                raise ValueError(f"Unknown model: {self.model}")
        else:
            raise TypeError(f"Expected str or RegisteredModel, got {type(self.model)}")

    def create(self) -> Runnable:
        """Agent Executor"""
        # For backwards compatibility
        return self()

    def __call__(self) -> Runnable:
        model = self._create_model()
        env = self.task.create_environment()

        model = _bind_tools(model, env.tools)

        if self.rate_limiter is not None:
            # Rate limited model
            model = rate_limiting.with_rate_limit(model, self.rate_limiter)

        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    self.task.instructions,
                ),
                ("user", "{input}"),
                MessagesPlaceholder(variable_name="agent_scratchpad"),
            ]
        )

        runnable_agent = (
            {
                "input": lambda x: x["input"],
                "agent_scratchpad": lambda x: format_to_openai_tool_messages(
                    x["intermediate_steps"]
                ),
            }
            | prompt
            | model
            | OpenAIToolsAgentOutputParser()
        )
        if self.num_retries > 0:
            runnable_agent = runnable_agent.with_retry(
                stop_after_attempt=self.num_retries + 1,
            )
        runnable = AgentExecutor(
            agent=runnable_agent,
            tools=env.tools,
            handle_parsing_errors=True,
            return_intermediate_steps=True,
        )

        # Returns `state` in the output if the environment has a state reader
        # makes sure that `output` is always in the output
        return apply_agent_executor_adapter(runnable, state_reader=env.read_state)



from langchain.smith.evaluation.config import RunEvalConfig, SingleKeyEvalConfig
from langsmith.evaluation.evaluator import (
    EvaluationResult,
    run_evaluator,
)
from langsmith.schemas import Example, Run

from langchain_benchmarks.extraction.tasks.chat_extraction.schema import GenerateTicket


@run_evaluator
def json_schema(run: Run, example: Example) -> EvaluationResult:
    """Evaluate the json schema of the generated ticket."""
    score, comment = None, None
    try:
        GenerateTicket.parse_obj(run.outputs["output"])
        score = 1
    except Exception as e:
        comment = repr(e)
        score = 0

    return EvaluationResult(
        key="json_schema",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_toxicity_similarity(run: Run, example: Example) -> EvaluationResult:
    """Evaluate the toxicity of the generated ticket."""
    gt = example.outputs["output"]["question"]["toxicity"]
    score, comment = None, None
    # Toxicity should be a on scale from 0 to 5
    try:
        pred = run.outputs["output"]["question"]["toxicity"]
        score = 1 - abs(gt - float(pred)) / 5
    except Exception as e:
        comment = repr(e)
        # Forgot to predict / mis-structured
        score = 0
    return EvaluationResult(
        key="toxicity_similarity",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_sentiment_similarity(run: Run, example: Example) -> EvaluationResult:
    """Evaluate the sentiment of the generated ticket."""
    gt = example.outputs["output"]["question"]["sentiment"]
    ordinal_map = {
        "negative": 0,
        "neutral": 1,
        "positive": 2,
    }
    gt_score = ordinal_map.get(str(gt).lower())
    score, comment = None, None
    # Sentiment is an enum, "Negative", "Neutral", "Positive"
    try:
        pred = run.outputs["output"]["question"]["sentiment"]
        pred_score = ordinal_map.get(str(pred).lower())
        score = 1 - (abs(gt_score - float(pred_score)) / 2)
    except Exception as e:
        comment = repr(e)
        # Forgot to predict / mis-structured
        score = 0
    return EvaluationResult(
        key="sentiment_similarity",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_confidence_level_similarity(
    run: Run, example: Example
) -> EvaluationResult:
    """Evaluate the confidence level of the generated ticket.
    This is a binary T/F question."""
    gt = example.outputs["output"]["response"]["confidence_level"]
    score, comment = None, None
    try:
        pred = run.outputs["output"]["response"]["confidence_level"]
        score = 1 - (abs(gt - float(pred)) / 5)
    except Exception as e:
        comment = repr(e)
        score = 0
    return EvaluationResult(
        key="confidence_level_similarity",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_question_category_similarity(
    run: Run, example: Example
) -> EvaluationResult:
    """Evaluate the question category of the generated ticket.
    This is a binary T/F question."""
    gt = example.outputs["output"]["question"]["question_category"]

    score, comment = None, None
    try:
        pred = run.outputs["output"]["question"]["question_category"]
        score = int(gt == pred)
    except Exception as e:
        comment = repr(e)
        # Forgot to predict / mis-structured
        score = 0
    return EvaluationResult(
        key="question_category",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_off_topic(run: Run, example: Example) -> EvaluationResult:
    """Evaluate the off topic of the generated ticket.
    This is a binary T/F question."""
    gt = example.outputs["output"]["question"]["is_off_topic"]
    score, comment = None, None
    try:
        pred = run.outputs["output"]["question"].get("is_off_topic")
        score = int(gt == pred)
    except Exception as e:
        comment = repr(e)
        # Forgot to predict / mis-structured
        score = 0
    return EvaluationResult(
        key="off_topic_similarity",
        score=score,
        comment=comment,
    )


@run_evaluator
def evaluate_programming_language(run: Run, example: Example) -> EvaluationResult:
    """Evaluate the programming language of the generated ticket.
    This is a binary T/F question."""
    gt = example.outputs["output"]["question"]["programming_language"]
    score, comment = None, None
    try:
        pred = run.outputs["output"]["question"]["programming_language"]
        score = int(gt == pred)
    except Exception as e:
        comment = repr(e)
        # Forgot to predict / mis-structured
        score = 0
    return EvaluationResult(
        key="programming_language_similarity",
        score=score,
        comment=comment,
    )


def get_eval_config() -> RunEvalConfig:
    """Get the evaluation configuration for the chat extraction task."""
    return RunEvalConfig(
        evaluators=[
            # General aggregate score
            SingleKeyEvalConfig(
                # input key is ignored.
                evaluator_type="json_edit_distance",
                input_key="question",
            )
        ],
        custom_evaluators=[
            json_schema,
            evaluate_toxicity_similarity,
            evaluate_sentiment_similarity,
            evaluate_confidence_level_similarity,
            evaluate_question_category_similarity,
            evaluate_off_topic,
            evaluate_programming_language,
        ],
    )



import logging
import os
from typing import Callable, Iterable, Optional

from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.schema.retriever import BaseRetriever
from langchain.vectorstores.chroma import Chroma

from langchain_benchmarks.rag.utils._downloading import (
    fetch_remote_file,
)
from langchain_benchmarks.rag.utils.indexing import (
    get_hyde_retriever,
    get_parent_document_retriever,
    get_vectorstore_retriever,
)

logger = logging.getLogger(__name__)
_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
# Stores the scraped documents from the langchain docs website, week of 2023-11-12
REMOTE_DOCS_FILE = "https://storage.googleapis.com/benchmarks-artifacts/langchain-docs-benchmarking/docs.parquet"
DOCS_FILE = os.path.join(_DIRECTORY, "db_docs/docs.parquet")

_DEFAULT_SEARCH_KWARGS = {"k": 6}


def load_docs_from_parquet(filename: Optional[str] = None) -> Iterable[Document]:
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "Please install pandas to use the langchain docs benchmarking task.\n"
            "pip install pandas"
        )
    if filename is None:
        filename = DOCS_FILE
    if not os.path.exists(filename):
        fetch_remote_file(REMOTE_DOCS_FILE, filename)
    df = pd.read_parquet(filename)
    docs_transformed = [Document(**row) for row in df.to_dict(orient="records")]
    for doc in docs_transformed:
        for k, v in doc.metadata.items():
            if v is None:
                doc.metadata[k] = ""
        if not doc.page_content.strip():
            continue
        yield doc


def _chroma_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transform_docs: Optional[Callable] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs_from_parquet()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-b-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_vectorstore_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="langchain-docs",
        transform_docs=transform_docs,
        transformation_name=transformation_name,
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
    )


def _chroma_parent_document_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs_from_parquet()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-b-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_parent_document_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="langchain-docs",
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
        transformation_name=transformation_name,
    )


def _chroma_hyde_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs_from_parquet()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-hd-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_hyde_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="langchain-docs",
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
        transformation_name=transformation_name,
    )


RETRIEVER_FACTORIES = {
    "basic": _chroma_retriever_factory,
    "parent-doc": _chroma_parent_document_retriever_factory,
    "hyde": _chroma_hyde_retriever_factory,
}



"""Factory for creating agents for the tool usage task."""
from typing import Union

from langchain.agents.agent import (
    AgentExecutor,
    BaseMultiActionAgent,
    BaseSingleActionAgent,
)
from langchain_core.runnables import Runnable

from langchain_benchmarks.schema import ToolUsageTask
from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter


class CustomRunnableAgentFactory:
    """A factory for creating tool using agents.

    A factory for agents that do not leverage any special JSON mode for
    function usage; instead all function invocation behavior is implemented solely
    through prompt engineering and parsing.
    """

    def __init__(
        self,
        task: ToolUsageTask,
        agent: Union[Runnable, BaseSingleActionAgent, BaseMultiActionAgent],
    ) -> None:
        """Create an agent factory for the given tool usage task.

        Note: The agent should not be stateful, as it will be reused across
        multiple runs.

        Args:
            task: The task to create an agent factory for
            agent: The agent to use
        """
        self.task = task
        self.agent = agent

    def __call__(self) -> Runnable:
        env = self.task.create_environment()
        executor = AgentExecutor(
            agent=self.agent,
            tools=env.tools,
            handle_parsing_errors=True,
            return_intermediate_steps=True,
        )

        return apply_agent_executor_adapter(executor, state_reader=env.read_state)



import logging
import os
from functools import partial
from typing import Callable, Iterable, List, Optional

from langchain.chat_models import ChatOpenAI
from langchain.indexes import SQLRecordManager, index
from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser
from langchain.prompts import ChatPromptTemplate
from langchain.pydantic_v1 import BaseModel
from langchain.retrievers.multi_vector import MultiVectorRetriever
from langchain.retrievers.parent_document_retriever import ParentDocumentRetriever
from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.schema.retriever import BaseRetriever
from langchain.schema.runnable import Runnable, RunnableConfig
from langchain.schema.storage import BaseStore
from langchain.schema.vectorstore import VectorStore
from langchain.storage import InMemoryStore
from langchain.text_splitter import RecursiveCharacterTextSplitter, TextSplitter
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)
_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
COLLECTION_NAME = "langchain-docs"
RECORD_MANAGER_DB_URL = (
    os.environ.get("RECORD_MANAGER_DB_URL")
    or f"sqlite:///{_DIRECTORY}_record_manager.sql"
)


def transform_docs_parent_child(
    documents: Iterable[Document],
    child_splitter: TextSplitter,
    docstore: BaseStore,
    id_key: str,
    *,
    parent_splitter: Optional[TextSplitter] = None,
) -> Iterable[Document]:
    """Transforms documents into child <-> parent documents."""
    if parent_splitter is not None:
        documents = parent_splitter.split_documents(documents)
    doc_ids = []
    for doc in documents:
        yield doc
        _id = doc.metadata[id_key]
        doc_ids.append((_id, doc))
        sub_docs = child_splitter.split_documents([doc])
        for _doc in sub_docs:
            _doc.metadata[id_key] = _id
            yield _doc
    docstore.mset(doc_ids)


def _default_hyde_embedder():
    class hypotheticalQuestions(BaseModel):
        """Write user queries that could be answered by the document."""

        questions: List[str]

    return (
        (
            {"doc": lambda x: x.page_content}
            # Only asking for 3 hypothetical questions, but this could be adjusted
            | ChatPromptTemplate.from_messages(
                [
                    (
                        "system",
                        "You are an AI creating an inverted index for"
                        " document retrieval. "
                        "Analyze the content of the following document and generate"
                        " relevant user queries "
                        "that would likely retrieve this document. Document content:"
                        "\n\n```document.txt\n{doc}\n```",
                    ),
                    (
                        "user",
                        "Based on the document's content, what specific technical queries"
                        " or questions"
                        " are users likely to search that this document can answer?",
                    ),
                ]
            )
            | ChatOpenAI(max_retries=0, model="gpt-4-1106-preview").bind_functions(
                functions=[hypotheticalQuestions],
                function_call="hypotheticalQuestions",
            )
            | JsonKeyOutputFunctionsParser(key_name="questions")
        )
        .with_retry(stop_after_attempt=3)
        .with_config(
            run_name="HyDE",
            metadata={"benchmark_environment": "langchain-docs"},
            tags=["langchain-benchmarks"],
        )
    )


def transform_docs_hyde(
    documents: Iterable[Document],
    docstore: BaseStore,
    id_key: str,
    *,
    query_generator: Optional[Runnable] = None,
    runnable_config: Optional[RunnableConfig] = None,
) -> Iterable[Document]:
    """Generates hypothetical document embeddings."""
    if query_generator is None:
        query_generator = _default_hyde_embedder()
        logger.info(f"Using default query generator\n{query_generator}")
    generator = query_generator or _default_hyde_embedder()
    docs = list(documents)
    questions = generator.batch(
        docs, runnable_config or {"max_concurrency": 5}, return_exceptions=True
    )
    doc_ids = []
    for doc, expansions in zip(documents, questions):
        yield doc
        if isinstance(expansions, BaseException):
            logger.error(
                f"Error generating questions for document {doc.metadata[id_key]}"
            )
            continue
        expansion_docs = [
            Document(page_content=s, metadata=doc.metadata) for s in expansions
        ]
        yield from expansion_docs
        doc_ids.append((doc.metadata[id_key], doc))
    docstore.mset(doc_ids)


def create_index(
    docs: Iterable[Document],
    embedding: Embeddings,
    vectorstore: VectorStore,
    collection_name: str,
    *,
    transform_docs: Optional[Callable] = None,
    transformation_name: Optional[str] = None,
):
    if transform_docs:
        if not transformation_name:
            raise ValueError(
                "If you provide a transform function, you must also provide a "
                "transformation name to use for the record manager."
            )
        transformed_docs = transform_docs(docs)
    else:
        transformed_docs = docs
    transformation_name = transformation_name or "raw"
    vectorstore_name = vectorstore.__class__.__name__
    embedding_name = embedding.__class__.__name__
    record_manager = SQLRecordManager(
        f"{vectorstore_name}/{collection_name}_{vectorstore_name}_{embedding_name}_{transformation_name}",
        db_url=RECORD_MANAGER_DB_URL,
    )
    record_manager.create_schema()
    return index(
        tqdm(transformed_docs),
        record_manager,
        vectorstore,
        cleanup="full",
        source_id_key="source",
    )


def get_vectorstore_retriever(
    docs: Iterable[Document],
    embedding: Embeddings,
    vectorstore: VectorStore,
    collection_name: str,
    *,
    transform_docs: Optional[Callable] = None,
    transformation_name: Optional[str] = None,
    search_kwargs: Optional[dict] = None,
) -> BaseRetriever:
    """Index the documents (with caching) and return a vector store retriever."""
    index_stats = create_index(
        docs,
        embedding,
        vectorstore,
        collection_name=collection_name,
        transform_docs=transform_docs,
        transformation_name=transformation_name,
    )
    logger.info(f"Index stats: {index_stats}")
    kwargs = search_kwargs or {}
    kwargs.setdefault("metadata", {}).setdefault(
        "benchmark_environment", "langchain-docs"
    )
    kwargs.setdefault("tags", []).append("langchain-benchmarks")
    return vectorstore.as_retriever(**kwargs)


def get_parent_document_retriever(
    docs: Iterable[Document],
    embedding: Embeddings,
    vectorstore: VectorStore,
    collection_name: str,
    *,
    child_splitter: Optional[TextSplitter] = None,
    transformation_name: Optional[str] = None,
    id_key: str = "source",
    docstore: Optional[BaseStore] = None,
    parent_splitter: Optional[TextSplitter] = None,
    search_kwargs: Optional[dict] = None,
):
    """Index the documents (with caching) and return a parent document retriever."""
    docstore = docstore or InMemoryStore()
    if child_splitter is None:
        child_splitter = RecursiveCharacterTextSplitter(
            chunk_size=4000, chunk_overlap=200
        )
        transformation_name = "parent-document-recursive-cs4k_ol200"
        logger.info(f"Using default child splitter:\n{child_splitter}")
    else:
        if transformation_name is None:
            raise ValueError(
                "If you provide a custom child splitter, you must also provide a "
                "transformation name to use for the record manager."
            )
    transformation = partial(
        transform_docs_parent_child,
        child_splitter=child_splitter,
        docstore=docstore,
        parent_splitter=parent_splitter,
        id_key=id_key,
    )
    index_stats = create_index(
        docs,
        embedding,
        vectorstore,
        collection_name=collection_name,
        transform_docs=transformation,
        transformation_name=transformation_name,
    )
    logger.info(f"Index stats: {index_stats}")

    return ParentDocumentRetriever(
        tags=["langchain-benchmarks"],
        metadata={"benchmark_environment": "langchain-docs"},
        vectorstore=vectorstore,
        docstore=docstore,
        child_splitter=child_splitter,
        parent_splitter=parent_splitter,
        search_kwargs=search_kwargs or {},
        id_key=id_key,
    )


def get_hyde_retriever(
    docs: Iterable[Document],
    embedding: Embeddings,
    vectorstore: VectorStore,
    collection_name: str,
    *,
    docstore: Optional[BaseStore] = None,
    query_generator: Optional[Runnable] = None,
    id_key: str = "source",
    search_kwargs: Optional[dict] = None,
    transformation_name: Optional[str] = None,
):
    """Index the documents (with caching) and return a parent document retriever."""
    docstore = docstore or InMemoryStore()
    if query_generator is not None and transformation_name is None:
        raise ValueError(
            "If you provide a custom query generator, you must also provide a "
            "transformation name to use for the record manager."
        )
    transformation_name = transformation_name or "HyDE"
    transformation = partial(
        transform_docs_hyde,
        docstore=docstore,
        query_generator=query_generator,
        id_key=id_key,
    )
    index_stats = create_index(
        docs,
        embedding,
        vectorstore,
        collection_name=collection_name,
        transform_docs=transformation,
        transformation_name=transformation_name,
    )
    logger.info(f"Index stats: {index_stats}")
    metadata = {
        "benchmark_environment": "langchain-docs",
        "retriever_stragegy": "HyDE",
        "embedding": embedding.__class__.__name__,
        "vectorstore": vectorstore.__class__.__name__,
    }

    return MultiVectorRetriever(
        vectorstore=vectorstore,
        docstore=docstore,
        id_key=id_key,
        search_kwargs=search_kwargs,
        metadata=metadata,
        tags=["langchain-benchmarks"],
    )



"""Copy the public dataset to your own langsmith tenant."""
import functools
import json
import logging
import threading
import urllib.parse
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union, cast
from uuid import UUID

from langchain.smith import RunEvalConfig
from langchain.smith.evaluation import runner_utils as eval_runner_utils
from langchain_core import runnables
from langchain_core.runnables import config as runnable_config
from langchain_core.tracers.root_listeners import RootListenersTracer
from langsmith import Client, EvaluationResult
from langsmith.evaluation.evaluator import EvaluationResults
from langsmith.schemas import DataType, Example, Run
from langsmith.utils import LangSmithNotFoundError
from tqdm import auto

logger = logging.getLogger(__name__)

API_URL = "https://api.smith.langchain.com/"


def _parse_token_or_url(url_or_token: str, api_url: str) -> Tuple[str, Optional[str]]:
    """Parse a public dataset URL or share token."""
    try:
        UUID(url_or_token)
        return api_url, url_or_token
    except ValueError:
        pass

    # Then it's a URL
    parsed_url = urllib.parse.urlparse(url_or_token)
    # Extract the UUID from the path
    path_parts = parsed_url.path.split("/")
    token_uuid = path_parts[-2] if len(path_parts) >= 2 else None
    return API_URL, token_uuid


# PUBLIC API


def clone_public_dataset(
    token_or_url: str,
    *,
    dataset_name: Optional[str] = None,
    source_api_url: str = API_URL,
) -> None:
    """Clone a public dataset to your own langsmith tenant.

    This operation is idempotent. If you already have a dataset with the given name,
    this function will do nothing.

    Args:
        token_or_url (str): The token of the public dataset to clone.
        dataset_name (str): The name of the dataset to create in your tenant.
        source_api_url: The URL of the langsmith server where the data is hosted:w
    """
    source_api_url, token_uuid = _parse_token_or_url(token_or_url, source_api_url)
    source_client = Client(api_url=source_api_url, api_key="placeholder")
    ds = source_client.read_shared_dataset(token_uuid)
    dataset_name = dataset_name or ds.name
    client = Client()  # Client used to write to langsmith
    try:
        dataset = client.read_dataset(dataset_name=dataset_name)

        if dataset:
            print(f"Dataset {dataset_name} already exists. Skipping.")
            print(f"You can access the dataset at {dataset.url}.")
            return
    except LangSmithNotFoundError:
        pass

    try:
        # Fetch examples first
        examples = auto.tqdm(list(source_client.list_shared_examples(token_uuid)))
        print("Finished fetching examples. Creating dataset...")
        dataset = client.create_dataset(dataset_name=dataset_name)
        print(f"New dataset created you can access it at {dataset.url}.")
        try:
            client.create_examples(
                inputs=[e.inputs for e in examples],
                outputs=[e.outputs for e in examples],
                dataset_id=dataset.id,
            )
        except BaseException as e:
            # Let's not do automatic clean up for now in case there might be
            # some other reasons why create_examples fails (i.e., not network issue or
            # keyboard interrupt).
            # The risk is that this is an existing dataset that has valid examples
            # populated from another source so we don't want to delete it.
            print(
                f"An error occurred while creating dataset {dataset_name}. "
                "You should delete it manually."
            )
            raise e

        print("Done creating dataset.")
    finally:
        del source_client
        del client


def download_public_dataset(
    token_or_url: str,
    *,
    path: Optional[Union[str, Path]] = None,
    api_url: str = API_URL,
) -> None:
    """Download a public dataset."""
    api_url, token_uuid = _parse_token_or_url(token_or_url, api_url)
    _path = str(path) if path else f"{token_uuid}.json"
    if not _path.endswith(".json"):
        raise ValueError(f"Path must end with .json got: {_path}")

    # This the client where the source data lives
    # The destination for the dataset is the local filesystem
    source_client = Client(api_url=api_url, api_key="placeholder")

    try:
        # Fetch examples first
        print("Fetching examples...")
        examples = auto.tqdm(list(source_client.list_shared_examples(token_uuid)))
        with open(str(_path), mode="w", encoding="utf-8") as f:
            jsonifable_examples = [json.loads(example.json()) for example in examples]
            json.dump(jsonifable_examples, f, indent=2)
        print("Done fetching examples.")
    finally:
        del source_client


def exists_public_dataset(token_or_url: str, *, api_url: str = API_URL) -> bool:
    """Check if a public dataset exists."""
    api_url, uuid = _parse_token_or_url(token_or_url, api_url)
    source_client = Client(api_url=api_url, api_key="placeholder")
    try:
        try:
            source_client.read_shared_dataset(uuid)
            return True
        except LangSmithNotFoundError:
            return False

    finally:
        del source_client


def _select_eval_results(
    results: Union[EvaluationResult, EvaluationResults],
) -> List[EvaluationResult]:
    if isinstance(results, EvaluationResult):
        results_ = [results]
    elif isinstance(results, dict) and "results" in results:
        results_ = cast(List[EvaluationResult], results["results"])
    else:
        raise TypeError(
            f"Invalid evaluation result type {type(results)}."
            " Expected EvaluationResult or EvaluationResults."
        )
    return results_


def _is_jupyter_environment() -> bool:
    try:
        from IPython import get_ipython

        res = get_ipython()
        return get_ipython() is not None and "zmqshell" in str(type(res))
    except ImportError:
        return False


def _display_aggregate_results(aggregate_results: Any) -> None:
    if _is_jupyter_environment():
        from IPython.display import HTML, display

        display(HTML("<h3>Experiment Results:</h3>"))
        display(aggregate_results)
    else:
        formatted_string = aggregate_results.to_string(
            float_format=lambda x: f"{x:.2f}", justify="right"
        )
        print("\n Experiment Results:")
        print(formatted_string)


def run_without_langsmith(
    path_or_token_id: Union[str, Path],
    llm_or_chain_factory: Union[
        Callable[[], runnables.Runnable], Callable[[dict], Any]
    ],
    *,
    evaluation: Optional[RunEvalConfig] = None,
    concurrency_level: int = 5,
    verbose: bool = True,
) -> None:
    """Run a public dataset without langsmith."""
    from langchain.smith.evaluation.runner_utils import (
        _setup_evaluation,
        _wrap_in_chain_factory,
    )

    if isinstance(path_or_token_id, Path) or path_or_token_id.endswith(".json"):
        dataset_path = path_or_token_id
    else:
        _, token_uuid = _parse_token_or_url(path_or_token_id, API_URL)
        dataset_path = f"{token_uuid}.json"
        if not Path(dataset_path).exists():
            download_public_dataset(path_or_token_id, path=dataset_path)
    if not dataset_path.endswith(".json"):
        raise ValueError(f"Unrecognized dataset path: {path_or_token_id}")
    with open(str(dataset_path), encoding="utf-8") as f:
        example_dicts = json.load(f)
    examples = [Example(**example_dict) for example_dict in example_dicts]
    wrapped_model = _wrap_in_chain_factory(llm_or_chain_factory)
    run_evaluators = _setup_evaluation(
        llm_or_chain_factory=wrapped_model,
        examples=examples,
        evaluation=evaluation,
        data_type=DataType.kv,
    )

    all_eval_results = {}
    results_lock = threading.RLock()
    _progress_bar = iter(
        auto.tqdm(
            iterable=range(len(examples)),
            desc="Running Evaluation",
            unit="example",
            total=len(examples),
        )
    )

    def _evaluate_run(run: Run, example: Example):
        with results_lock:
            next(_progress_bar)
            example_result = all_eval_results.setdefault(str(example.id), {}) or {}
            example_result.update(
                {
                    "input": run.inputs,
                    "execution_time": (
                        (run.end_time - run.start_time).total_seconds()
                        if run.end_time
                        else None
                    ),
                    "run_id": str(run.id),
                }
            )
            if run.error is not None:
                example_result["Error"] = run.error
            else:
                example_result["output"] = run.outputs
            all_eval_results[str(example.id)] = example_result
        if run_evaluators is None:
            return
        feedback = []
        for evaluator in run_evaluators:
            try:
                eval_results = evaluator.evaluate_run(run, example)
            except Exception as e:
                logger.error(f"Failed to evaluate run {run.id}: {repr(e)}")
                continue
            flattened = _select_eval_results(eval_results)
            feedback.extend(flattened)

        with results_lock:
            example_result = all_eval_results.setdefault(str(example.id), {}) or {}
            example_result.update(
                {
                    "feedback": feedback,
                }
            )
            all_eval_results[str(example.id)] = example_result

    configs = [
        runnable_config.RunnableConfig(
            callbacks=[
                RootListenersTracer(
                    config={},
                    on_start=None,
                    on_end=functools.partial(_evaluate_run, example=example),
                    on_error=functools.partial(_evaluate_run, example=example),
                ),
            ],
            max_concurrency=concurrency_level,
        )
        for example in examples
    ]

    def run_runnable(x: dict) -> Any:
        model = wrapped_model()
        return model.invoke(x)

    runnables.RunnableLambda(run_runnable).batch(
        inputs=[example.inputs for example in examples],
        config=configs,
        return_exceptions=True,
    )
    results = eval_runner_utils.TestResult(
        project_name="Local",
        results=all_eval_results,
    )
    if verbose:
        try:
            agg_feedback = results.get_aggregate_feedback()
            _display_aggregate_results(agg_feedback)
        except Exception as e:
            logger.debug(f"Failed to print aggregate feedback: {repr(e)}")
    return results



"""Package for helping to evaluate agent runs."""
from langchain_benchmarks.tool_usage.agents import apply_agent_executor_adapter
from langchain_benchmarks.tool_usage.evaluators import get_eval_config

# Please keep this list sorted!
__all__ = [
    "apply_agent_executor_adapter",
    "get_eval_config",
]



from typing import Optional

from langchain.chat_models import ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.smith import RunEvalConfig


def get_eval_config(eval_llm: Optional[BaseChatModel] = None) -> RunEvalConfig:
    eval_llm = eval_llm or ChatOpenAI(
        model="gpt-4",
        temperature=0,
        model_kwargs={"seed": 42},
        max_retries=1,
        request_timeout=60,
    )
    """Get the evaluation configuration for the email task."""
    return RunEvalConfig(
        evaluators=[
            "json_edit_distance",
            RunEvalConfig.LabeledScoreString(
                criteria={
                    "accuracy": """
    Score 1: The answer is incorrect and unrelated to the question or reference document.
    Score 3: The answer is partially correct but has more than one omission or major errors.
    Score 5: The answer is mostly correct but has more than one omission or major error.
    Score 7: The answer is mostly correct but has at most one omission or major error.
    Score 9: The answer is mostly correct with no omissions and only minor errors, and aligns with the reference document.
    Score 10: The answer is correct, complete, and aligns with the reference document. Extra information is acceptable if it is sensible.

    If the reference answer contains multiple alternatives, the predicted answer must only match one of the alternatives to be considered correct.
    If the predicted answer contains additional helpful and accurate information that is not present in the reference answer, it should still be considered correct and not be penalized.
    """  # noqa
                },
                llm=eval_llm,
                normalize_by=10.0,
            ),
        ],
    )



from langchain_benchmarks.tool_usage.agents.adapters import apply_agent_executor_adapter
from langchain_benchmarks.tool_usage.agents.anthropic_tool_user import (
    AnthropicToolUserFactory,
)
from langchain_benchmarks.tool_usage.agents.experimental.factory import (
    CustomAgentFactory,
)
from langchain_benchmarks.tool_usage.agents.openai_assistant import (
    OpenAIAssistantFactory,
)
from langchain_benchmarks.tool_usage.agents.openai_functions import OpenAIAgentFactory
from langchain_benchmarks.tool_usage.agents.runnable_agent import (
    CustomRunnableAgentFactory,
)

__all__ = [
    "OpenAIAgentFactory",
    "OpenAIAssistantFactory",
    "apply_agent_executor_adapter",
    "CustomAgentFactory",
    "AnthropicToolUserFactory",
    "CustomRunnableAgentFactory",
]



AGENT_INSTRUCTIONS_XML_FORMAT = """\
In this environment you have access to a set of tools you can use to answer the user's question.

You may call them like this:
<function_calls>
<invoke>
<tool_name>$TOOL_NAME</tool_name>
<parameters>
<$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>
...
</parameters>
</invoke>
</function_calls>

Here are the tools available:

{tool_description}
"""  # noqa: E501

_AGENT_INSTRUCTIONS_BLOB_STYLE = """\
In this environment you have access to a set of tools you can use to answer the user's question.

Here are the tools available:

{tool_description}

You may call one tool at a time using a format that includes <tool> and </tool> tag. 

Inside the tag the content is a python dictionary that uses python literals (e.g., numbers, strings, lists, dictionaries, etc.) to specify the tool invocation.

It must match the schema of the function as described in the tool description.
"arguments" is a dictionary of the arguments to the function.

<tool>
{{
    "tool_name": $TOOL_NAME,
    "arguments": $ARGUMENTS
}}
</tool>

If you do not know the answer use more tools. You can only take a single action at a time.\
"""  # noqa: E501



from importlib import metadata

from langchain_benchmarks.model_registration import model_registry
from langchain_benchmarks.rate_limiting import RateLimiter
from langchain_benchmarks.registration import registry
from langchain_benchmarks.utils._langsmith import (
    clone_public_dataset,
    download_public_dataset,
)

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)


# Please keep this list sorted!
__all__ = [
    "__version__",
    "clone_public_dataset",
    "download_public_dataset",
    "model_registry",
    "RateLimiter",
    "registry",
]



"""Registry of environments for ease of access."""

from langchain_benchmarks.extraction.tasks import chat_extraction, email_task
from langchain_benchmarks.rag.tasks import (
    LANGCHAIN_DOCS_TASK,
    MULTI_MODAL_SLIDE_DECKS_TASK,
    SEMI_STRUCTURED_REPORTS_TASK,
)
from langchain_benchmarks.schema import Registry
from langchain_benchmarks.tool_usage.tasks import (
    multiverse_math,
    relational_data,
    type_writer,
    type_writer_26_funcs,
)

# Using lower case naming to make a bit prettier API when used in a notebook
registry = Registry(
    tasks=[
        type_writer.TYPE_WRITER_TASK,
        type_writer_26_funcs.TYPE_WRITER_26_FUNCS_TASK,
        relational_data.RELATIONAL_DATA_TASK,
        multiverse_math.MULTIVERSE_MATH,
        multiverse_math.MULTIVERSE_MATH_TINY,
        email_task.EMAIL_EXTRACTION_TASK,
        chat_extraction.CHAT_EXTRACTION_TASK,
        LANGCHAIN_DOCS_TASK,
        SEMI_STRUCTURED_REPORTS_TASK,
        MULTI_MODAL_SLIDE_DECKS_TASK,
    ]
)



from typing import Optional

from langchain.chat_models import ChatOpenAI
from langchain.evaluation import load_evaluator
from langchain.smith import RunEvalConfig

try:
    from langchain.schema.language_model import BaseLanguageModel
except ImportError:
    from langchain_core.language_models import BaseLanguageModel
from langsmith.evaluation.evaluator import EvaluationResult, RunEvaluator
from langsmith.schemas import Example, Run


# TODO: Split this into an assertion-by-assertion evaluator
# TODO: Combine with a document relevance evaluator (to report retriever performance)
class FaithfulnessEvaluator(RunEvaluator):
    def __init__(self, llm: Optional[BaseLanguageModel] = None):
        self.evaluator = load_evaluator(
            "labeled_score_string",
            criteria={
                "faithfulness": """
Score 1: The answer directly contradicts the information provided in the reference docs.
Score 3: The answer contains a mix of correct information from the reference docs and incorrect or unverifiable information not found in the docs.
Score 5: The answer is mostly aligned with the reference docs but includes extra information that, while not contradictory, is not verified by the docs.
Score 7: The answer aligns well with the reference docs but includes minor, commonly accepted facts not found in the docs.
Score 10: The answer perfectly aligns with and is fully entailed by the reference docs, with no extra information."""
            },
            llm=llm,
            normalize_by=10,
        )

    @staticmethod
    def _get_retrieved_docs(run: Run) -> str:
        # This assumes there is only one retriever in your chain.
        # To select more precisely, name your retrieval chain
        # using with_config(name="my_unique_name") and look up
        # by run.name
        runs = [run]
        while runs:
            run = runs.pop()
            if run.run_type == "retriever":
                return str(run.outputs["documents"])
            if run.child_runs:
                runs.extend(run.child_runs[::-1])
        return ""

    def evaluate_run(
        self, run: Run, example: Optional[Example] = None
    ) -> EvaluationResult:
        try:
            docs_string = self._get_retrieved_docs(run)
            docs_string = f"Reference docs:\n<DOCS>\n{docs_string}\n</DOCS>\n\n"
            input_query = run.inputs["question"]
            if run.outputs is not None and len(run.outputs) == 1:
                prediction = next(iter(run.outputs.values()))
            else:
                prediction = run.outputs["output"]
            result = self.evaluator.evaluate_strings(
                input=input_query,
                prediction=prediction,
                reference=docs_string,
            )
            return EvaluationResult(
                **{"key": "faithfulness", "comment": result.get("reasoning"), **result}
            )
        except Exception as e:
            return EvaluationResult(key="faithfulness", score=None, comment=repr(e))


_ACCURACY_CRITERION = {
    "accuracy": """
Score 1: The answer is incorrect and unrelated to the question or reference document.
Score 3: The answer shows slight relevance to the question or reference document but is largely incorrect.
Score 5: The answer is partially correct but has significant errors or omissions.
Score 7: The answer is mostly correct with minor errors or omissions, and aligns with the reference document.
Score 10: The answer is correct, complete, and perfectly aligns with the reference document.

If the reference answer contains multiple alternatives, the predicted answer must only match one of the alternatives to be considered correct.
If the predicted answer contains additional helpful and accurate information that is not present in the reference answer, it should still be considered correct.
"""  # noqa
}


def get_eval_config() -> RunEvalConfig:
    """Returns the evaluator for the environment."""
    eval_llm = ChatOpenAI(
        model="gpt-4",
        temperature=0.0,
        model_kwargs={"seed": 42},
        max_retries=1,
        request_timeout=60,
    )
    # Use a longer-context LLM to check documents
    faithfulness_eval_llm = ChatOpenAI(
        model="gpt-4-1106-preview",
        temperature=0.0,
        model_kwargs={"seed": 42},
        max_retries=1,
        request_timeout=60,
    )

    return RunEvalConfig(
        evaluators=[
            RunEvalConfig.LabeledScoreString(
                criteria=_ACCURACY_CRITERION, llm=eval_llm, normalize_by=10.0
            ),
            RunEvalConfig.EmbeddingDistance(),
        ],
        custom_evaluators=[FaithfulnessEvaluator(llm=faithfulness_eval_llm)],
    )



import os

import requests


def is_folder_populated(folder: str):
    if os.path.exists(folder):
        return any(os.scandir(folder))
    return False


def fetch_remote_file(remote: str, local: str):
    if not os.path.exists(local):
        print(f"File {local} does not exist. Downloading from GCS...")
        if not os.path.exists(os.path.dirname(local)):
            os.makedirs(os.path.dirname(local))
        r = requests.get(remote, allow_redirects=True)
        with open(local, "wb") as f:
            f.write(r.content)
        print(f"File {remote} downloaded.")



from typing import Iterable

from langchain.schema.document import Document

from langchain_benchmarks.rag.tasks.langchain_docs import architectures, indexing
from langchain_benchmarks.rag.tasks.langchain_docs.indexing.retriever_registry import (
    DOCS_FILE,
    load_docs_from_parquet,
)
from langchain_benchmarks.schema import RetrievalTask

# URL of public LangChain Docs dataset
DATASET_ID = "https://smith.langchain.com/public/452ccafc-18e1-4314-885b-edd735f17b9d/d"


def load_cached_docs() -> Iterable[Document]:
    """Load the docs from the cached file."""
    return load_docs_from_parquet(DOCS_FILE)


LANGCHAIN_DOCS_TASK = RetrievalTask(
    name="LangChain Docs Q&A",
    dataset_id=DATASET_ID,
    retriever_factories=indexing.RETRIEVER_FACTORIES,
    architecture_factories=architectures.ARCH_FACTORIES,
    get_docs=load_cached_docs,
    description=(
        """\
Questions and answers based on a snapshot of the LangChain python docs.

The environment provides the documents and the retriever information.

Each example is composed of a question and reference answer.

Success is measured based on the accuracy of the answer relative to the reference answer.
We also measure the faithfulness of the model's response relative to the retrieved documents (if any).
"""  # noqa: E501
    ),
)



from enum import Enum
from typing import List, Optional

from langchain.pydantic_v1 import BaseModel, Field


class QuestionCategory(str, Enum):
    IMPLEMENTATION_ISSUES = "Implementation Issues"  # about existing implementation
    FEATURE_REQUESTS = "Feature Requests"
    CONCEPT_EXPLANATIONS = "Concept Explanations"
    CODE_OPTIMIZATION = "Code Optimization"
    SECURITY_AND_PRIVACY_CONCERNS = "Security and Privacy Concerns"
    MODEL_TRAINING_AND_FINE_TUNING = "Model Training and Fine-tuning"
    DATA_HANDLING_AND_MANIPULATION = "Data Handling and Manipulation"
    USER_INTERACTION_FLOW = "User Interaction Flow"
    TECHNICAL_INTEGRATION = "Technical Integration"
    ERROR_HANDLING_AND_LOGGING = "Error Handling and Logging"
    CUSTOMIZATION_AND_CONFIGURATION = "Customization and Configuration"
    EXTERNAL_API_AND_DATA_SOURCE_INTEGRATION = (
        "External API and Data Source Integration"
    )
    LANGUAGE_AND_LOCALIZATION = "Language and Localization"
    STREAMING_AND_REAL_TIME_PROCESSING = "Streaming and Real-time Processing"
    TOOL_DEVELOPMENT = "Tool Development"
    FUNCTION_CALLING = "Function Calling"
    LLM_INTEGRATIONS = "LLM Integrations"
    GENERAL_AGENT_QUESTIONS = "General Agent Question"
    GENERAL_CHIT_CHAT = "General Chit Chat"
    MEMORY = "Memory"
    DEBUGGING_HELP = "Debugging Help"
    APPLICATION_DESIGN = "Application Design"
    PROMPT_TEMPLATES = "Prompt Templates"
    COST_TRACKING = "Cost Tracking"
    OTHER = "Other"


class Sentiment(str, Enum):
    NEGATIVE = "Negative"
    NEUTRAL = "Neutral"
    POSITIVE = "Positive"


class ProgrammingLanguage(str, Enum):
    PYTHON = "python"
    JAVASCRIPT = "javascript"
    TYPESCRIPT = "typescript"
    UNKNOWN = "unknown"
    OTHER = "other"


class QuestionCategorization(BaseModel):
    question_category: QuestionCategory
    category_if_other: Optional[str] = Field(
        default=None, description="question category if the category above is 'other'"
    )
    is_off_topic: bool = Field(
        description="If the input is general chit chat or does not pertain to technical inqueries about LangChain or building/debugging applications with LLMs/AI, it is off topic. For context, LangChain is a library and framework designed"
        " to assist in building applications with LLMs. Questions may also be about similar packages like LangServe, LangSmith, OpenAI, Anthropic, vectorstores, agents, etc."
    )
    toxicity: int = Field(
        ge=0, lt=6, description="Whether or not the input question is toxic"
    )
    sentiment: Sentiment
    programming_language: ProgrammingLanguage


#  resolve the issue, provide guidance, or ask for more information
class ResponseType(str, Enum):
    RESOLVE_ISSUE = "resolve issue"
    PROVIDE_GUIDANCE = "provide guidance"
    REQUEST_INFORMATION = "request information"
    GIVE_UP = "give up"
    NONE = "none"
    OTHER = "other"


class ResponseCategorization(BaseModel):
    response_type: ResponseType
    response_type_if_other: Optional[str] = None
    confidence_level: int = Field(
        ge=0, lt=6, description="The confidence of the assistant in its answer."
    )
    followup_actions: Optional[List[str]] = Field(
        description="Actions the assistant recommended the user take."
    )


class GenerateTicket(BaseModel):
    """Generate a ticket containing all the extracted information."""

    issue_summary: str = Field(
        description="short (<10 word) summary of the issue or question"
    )
    question: QuestionCategorization = Field(
        description="Information inferred from the the question."
    )
    response: ResponseCategorization = Field(
        description="Information inferred from the the response."
    )



import logging
import os
import zipfile
from pathlib import Path
from typing import Callable, Iterable, Optional

from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.schema.retriever import BaseRetriever
from langchain.vectorstores.chroma import Chroma

from langchain_benchmarks.rag.utils._downloading import (
    fetch_remote_file,
    is_folder_populated,
)
from langchain_benchmarks.rag.utils.indexing import (
    get_hyde_retriever,
    get_parent_document_retriever,
    get_vectorstore_retriever,
)

logger = logging.getLogger(__name__)
_DIRECTORY = Path(os.path.abspath(__file__)).parent
# Stores the zipped pdfs for this dataset
REMOTE_DOCS_FILE = "https://storage.googleapis.com/benchmarks-artifacts/langchain-docs-benchmarking/semi_structured_earnings.zip"
DOCS_DIR = _DIRECTORY / "pdfs"

_DEFAULT_SEARCH_KWARGS = {"k": 6}


def fetch_raw_docs(
    filename: Optional[str] = None, docs_dir: Optional[str] = None
) -> None:
    filename = filename or _DIRECTORY / Path(REMOTE_DOCS_FILE).name
    docs_dir = docs_dir or DOCS_DIR
    if not is_folder_populated(docs_dir):
        fetch_remote_file(REMOTE_DOCS_FILE, filename)
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(docs_dir)

        os.remove(filename)


def get_file_names() -> Iterable[Path]:
    fetch_raw_docs()
    # Traverse the directory and partition the pdfs
    for path in DOCS_DIR.glob("*.pdf"):
        yield path


def partition_pdfs(path: Path, *, config: Optional[dict] = None):
    try:
        from unstructured.partition.pdf import partition_pdf
    except ImportError:
        raise ImportError(
            "Please install the unstructured package to use this example.\n"
            "pip install unstructured"
        )

    config = {
        # Unstructured first finds embedded image blocks
        "extract_images_in_pdf": False,
        # Use layout model (YOLOX) to get bounding boxes (for tables) and find titles
        # Titles are any sub-section of the document
        "infer_table_structure": True,
        # Post processing to aggregate text once we have the title
        "chunking_strategy": "by_title",
        # Chunking params to aggregate text blocks
        # Attempt to create a new chunk 3800 chars
        # Attempt to keep chunks > 2000 chars
        "max_characters": 4000,
        "new_after_n_chars": 3800,
        "combine_text_under_n_chars": 2000,
        **(config or {}),
    }
    raw_pdf_elements = partition_pdf(
        filename=str(path), image_output_dir_path=str(path), **config
    )

    # Categorize by type
    for element in raw_pdf_elements:
        if "unstructured.documents.elements.Table" in str(type(element)):
            yield Document(
                page_content=str(element),
                metadata={"element_type": "table", "source": str(path.name)},
                id=str(element),
            )
        elif "unstructured.documents.elements.CompositeElement" in str(type(element)):
            yield Document(
                page_content=str(element),
                metadata={"element_type": "composite", "source": str(path.name)},
                id=str(element),
            )
        else:
            logger.debug(f"Skipping element of type {type(element)}")


def load_docs(*, unstructured_config: Optional[dict] = None) -> Iterable[Document]:
    for path in get_file_names():
        yield from partition_pdfs(path, config=unstructured_config)


def _chroma_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transform_docs: Optional[Callable] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-ss-b-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_vectorstore_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="semi-structured-earnings-b",
        transform_docs=transform_docs,
        transformation_name=transformation_name,
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
    )


def _chroma_parent_document_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-ss-pd-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_parent_document_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="semi-structured-earnings-pd",
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
        transformation_name=transformation_name,
    )


def _chroma_hyde_retriever_factory(
    embedding: Embeddings,
    *,
    docs: Optional[Iterable[Document]] = None,
    search_kwargs: Optional[dict] = None,
    transformation_name: Optional[str] = None,
) -> BaseRetriever:
    docs = docs or load_docs()
    embedding_name = embedding.__class__.__name__
    vectorstore = Chroma(
        collection_name=f"lcbm-ss-hd-{embedding_name}-{transformation_name}",
        embedding_function=embedding,
        persist_directory="./chromadb",
    )
    return get_hyde_retriever(
        docs,
        embedding,
        vectorstore,
        collection_name="semi-structured-earnings-hd",
        search_kwargs=search_kwargs or _DEFAULT_SEARCH_KWARGS,
        transformation_name=transformation_name,
    )


RETRIEVER_FACTORIES = {
    "basic": _chroma_retriever_factory,
    "parent-doc": _chroma_parent_document_retriever_factory,
    "hyde": _chroma_hyde_retriever_factory,
}



from enum import Enum
from typing import List, Optional

from langchain.prompts import ChatPromptTemplate
from langchain.pydantic_v1 import BaseModel, Field

from langchain_benchmarks.schema import ExtractionTask


class ToneEnum(str, Enum):
    """The tone of the email."""

    positive = "positive"
    negative = "negative"


class Email(BaseModel):
    """Relevant information about an email."""

    sender: Optional[str] = Field(None, description="The sender's name, if available")
    sender_phone_number: Optional[str] = Field(
        None, description="The sender's phone number, if available"
    )
    sender_address: Optional[str] = Field(
        None, description="The sender's address, if available"
    )
    action_items: List[str] = Field(
        ..., description="A list of action items requested by the email"
    )
    topic: str = Field(
        ..., description="High level description of what the email is about"
    )
    tone: ToneEnum = Field(..., description="The tone of the email.")


# This is a default prompt that works for chat models.
DEFAULT_CHAT_MODEL_PROMPT = ChatPromptTemplate.from_messages(
    [
        ("system", "You are an expert researcher."),
        (
            "human",
            "What can you tell me about the following email? Make sure to "
            "extract the question in the correct format. "
            "Here is the email:\n ```\n{input}\n```",
        ),
    ]
)

EMAIL_EXTRACTION_TASK = ExtractionTask(
    name="Email Extraction",
    dataset_id="https://smith.langchain.com/public/a1742786-bde5-4f51-a1d8-e148e5251ddb/d",
    schema=Email,
    description="""\
A dataset of 42 real emails deduped from a spam folder, with semantic HTML tags removed, \
as well as a script for initial extraction and formatting of other emails from \
an arbitrary .mbox file like the one exported by Gmail.

Some additional cleanup of the data was done by hand after the initial pass.

See https://github.com/jacoblee93/oss-model-extraction-evals.
    """,
    instructions=DEFAULT_CHAT_MODEL_PROMPT,
)
